@@ -58,7 +58,7 @@ def heredoc(s):
 
 motd = heredoc('''
 
-    This is the Toil appliance. You can run your Toil script directly on the appliance. 
+    This is the Toil appliance. You can run your Toil script directly on the appliance.
     Run toil <workflow>.py --help to see all options for running your workflow.
     For more information see http://toil.readthedocs.io/en/latest/
 
@@ -84,7 +84,7 @@ print(heredoc('''
         && apt-key adv --keyserver keyserver.ubuntu.com --recv 68576280
 
     RUN add-apt-repository -y ppa:deadsnakes/ppa
-    
+
     RUN apt-get -y update --fix-missing && \
         DEBIAN_FRONTEND=noninteractive apt-get -y upgrade && \
         DEBIAN_FRONTEND=noninteractive apt-get -y install {dependencies} && \
@@ -96,7 +96,9 @@ print(heredoc('''
         rm go1.13.3.linux-amd64.tar.gz && \
         mv go/bin/* /usr/bin/ && \
         mv go /usr/local/
-        
+
+    # Build Singularity, but only keep the binaries and scrap the GOPATH to
+    # save space. Hide its binary so we can wrap it.
     RUN mkdir -p $(go env GOPATH)/src/github.com/sylabs && \
         cd $(go env GOPATH)/src/github.com/sylabs && \
         git clone https://github.com/sylabs/singularity.git && \
@@ -105,22 +107,27 @@ print(heredoc('''
         ./mconfig && \
         cd ./builddir && \
         make -j4 && \
-        make install
-    
+        make install && \
+        cd && \
+        rm -Rf $(go env GOPATH) \
+        && mkdir -p /usr/local/libexec/toil && \
+        mv /usr/local/bin/singularity /usr/local/libexec/toil/singularity-real
     RUN mkdir /root/.ssh && \
         chmod 700 /root/.ssh
 
     ADD waitForKey.sh /usr/bin/waitForKey.sh
 
     ADD customDockerInit.sh /usr/bin/customDockerInit.sh
-
-    RUN chmod 777 /usr/bin/waitForKey.sh && chmod 777 /usr/bin/customDockerInit.sh
     
+    ADD singularity-wrapper.sh /usr/local/bin/singularity
+
+    RUN chmod 777 /usr/bin/waitForKey.sh && chmod 777 /usr/bin/customDockerInit.sh && chmod 777 /usr/local/bin/singularity
+
     # fixes an incompatibility updating pip on Ubuntu 16 w/ python3.8
     RUN sed -i "s/platform.linux_distribution()/('Ubuntu', '16.04', 'xenial')/g" /usr/lib/python3/dist-packages/pip/download.py
-    
+
     # The stock pip is too old and can't install from sdist with extras
-    RUN {pip} install --upgrade pip==20.0.2
+    RUN {pip} install --upgrade pip==21.0.1
 
     # Default setuptools is too old
     RUN {pip} install --upgrade setuptools==45
@@ -156,17 +163,24 @@ print(heredoc('''
     ENV TOIL_APPLIANCE_SELF {applianceSelf}
 
     RUN mkdir /var/lib/toil
-
     ENV TOIL_WORKDIR /var/lib/toil
 
+    # We want to get binaries mounted in from the environemnt on Toil-managed Kubernetes
+    env PATH /opt/bin:/usr/local/sbin:/usr/local/bin:/usr/sbin:/usr/bin:/sbin:/bin
+
+    # We want to pick the right Python when the user runs it
+    RUN rm /usr/bin/python3 && rm /usr/bin/python && \
+        ln -s /usr/bin/{python} /usr/bin/python3 && \
+        ln -s /usr/bin/python3 /usr/bin/python
+
     # This component changes most frequently and keeping it last maximizes Docker cache hits.
     COPY {sdistName} .
     RUN {pip} install {sdistName}[all]
     RUN rm {sdistName}
 
     # We intentionally inherit the default ENTRYPOINT and CMD from the base image, to the effect
-    # that the running appliance just gives you a shell. To start the Mesos master or slave
-    # daemons, the user # should override the entrypoint via --entrypoint.
+    # that the running appliance just gives you a shell. To start the Mesos daemons, the user
+    # should override the entrypoint via --entrypoint.
 
     RUN echo '[ ! -z "$TERM" -a -r /etc/motd ] && cat /etc/motd' >> /etc/bash.bashrc \
         && printf '{motd}' > /etc/motd
