@@ -16,13 +16,12 @@ import logging
 import os
 import shutil
 from abc import ABC, abstractmethod
-from collections import namedtuple
 from contextlib import contextmanager
-from typing import Any, Optional, Tuple
+from typing import Any, Optional, Tuple, Union, Dict, NamedTuple
 
 from toil.batchSystems.registry import (BATCH_SYSTEM_FACTORY_REGISTRY,
                                         DEFAULT_BATCH_SYSTEM)
-from toil.common import Toil, cacheDirName
+from toil.common import Toil, cacheDirName, Config
 from toil.deferred import DeferredFunctionManager
 from toil.fileStores.abstractFileStore import AbstractFileStore
 from toil.lib.threading import LastProcessStandingArena
@@ -33,36 +32,34 @@ except ImportError:
     # CWL extra not installed
     CWL_INTERNAL_JOBS = ()
 
+# Value to use as exitStatus in UpdatedBatchJobInfo.exitStatus when status is not available.
+EXIT_STATUS_UNAVAILABLE_VALUE = 255
 logger = logging.getLogger(__name__)
 
-UpdatedBatchJobInfo = namedtuple('UpdatedBatchJobInfo', (
-    'jobID',
-    # The exit status (integer value) of the job. 0 implies successful.
-    # EXIT_STATUS_UNAVAILABLE_VALUE is used when the exit status is not available (e.g. job is lost).
-    'exitStatus',
-    'exitReason',  # The exit reason, if available. One of BatchJobExitReason enum.
-    'wallTime'))
+
+UpdatedBatchJobInfo = NamedTuple('UpdatedBatchJobInfo',
+    [('jobID', str),
+     # The exit status (integer value) of the job. 0 implies successful.
+     # EXIT_STATUS_UNAVAILABLE_VALUE is used when the exit status is not available (e.g. job is lost).
+     ('exitStatus', int),
+     ('exitReason', Union[int, None]),  # The exit reason, if available. One of BatchJobExitReason enum.
+     ('wallTime', Union[float, int, None])])
+
+
+# Information required for worker cleanup on shutdown of the batch system.
+WorkerCleanupInfo = NamedTuple('WorkerCleanupInfo',
+    [('workDir', str),  # workdir path (where the cache would go)
+     ('workflowID', int),  # used to identify files specific to this workflow
+     ('cleanWorkDir', bool)])
 
 
 class BatchJobExitReason(enum.Enum):
-    FINISHED = 1  # Successfully finished.
-    FAILED = 2  # Job finished, but failed.
-    LOST = 3  # Preemptable failure (job's executing host went away).
-    KILLED = 4  # Job killed before finishing.
-    ERROR = 5  # Internal error.
-    MEMLIMIT = 6 # Job hit batch system imposed memory limit
-
-# Value to use as exitStatus in UpdatedBatchJobInfo.exitStatus when status is not available.
-EXIT_STATUS_UNAVAILABLE_VALUE = 255
-
-# A class containing the information required for worker cleanup on shutdown of the batch system.
-WorkerCleanupInfo = namedtuple('WorkerCleanupInfo', (
-    # A path to the value of config.workDir (where the cache would go)
-    'workDir',
-    # The value of config.workflowID (used to identify files specific to this workflow)
-    'workflowID',
-    # The value of the cleanWorkDir flag
-    'cleanWorkDir'))
+    FINISHED: int = 1  # Successfully finished.
+    FAILED: int = 2  # Job finished, but failed.
+    LOST: int = 3  # Preemptable failure (job's executing host went away).
+    KILLED: int = 4  # Job killed before finishing.
+    ERROR: int = 5  # Internal error.
+    MEMLIMIT: int = 6  # Job hit batch system imposed memory limit
 
 
 class AbstractBatchSystem(ABC):
@@ -113,11 +110,13 @@ class AbstractBatchSystem(ABC):
         raise NotImplementedError()
 
     @abstractmethod
-    def issueBatchJob(self, jobDesc):
+    def issueBatchJob(self, jobDesc, job_environment: Optional[Dict[str, str]] = None):
         """
         Issues a job with the specified command to the batch system and returns a unique jobID.
 
         :param jobDesc a toil.job.JobDescription
+        :param job_environment: a collection of job-specific environment variables
+                                to be set on the worker.
 
         :return: a unique jobID that can be used to reference the newly issued job
         :rtype: int
@@ -168,7 +167,7 @@ class AbstractBatchSystem(ABC):
         Returns information about job that has updated its status (i.e. ceased
         running, either successfully or with an error). Each such job will be
         returned exactly once.
-        
+
         Does not return info for jobs killed by killBatchJobs, although they
         may cause None to be returned earlier than maxWait.
 
@@ -176,28 +175,28 @@ class AbstractBatchSystem(ABC):
 
         :rtype: UpdatedBatchJobInfo or None
         :return: If a result is available, returns UpdatedBatchJobInfo.
-                 Otherwise it returns None. wallTime is the number of seconds (a strictly 
+                 Otherwise it returns None. wallTime is the number of seconds (a strictly
                  positive float) in wall-clock time the job ran for, or None if this
                  batch system does not support tracking wall time.
         """
         raise NotImplementedError()
-        
+
     def getSchedulingStatusMessage(self):
         """
         Get a log message fragment for the user about anything that might be
         going wrong in the batch system, if available.
-        
+
         If no useful message is available, return None.
-        
+
         This can be used to report what resource is the limiting factor when
         scheduling jobs, for example. If the leader thinks the workflow is
         stuck, the message can be displayed to the user to help them diagnose
         why it might be stuck.
-        
+
         :rtype: str or None
         :return: User-directed message about scheduling state.
         """
-        
+
         # Default implementation returns None.
         # Override to provide scheduling status information.
         return None
@@ -233,17 +232,17 @@ class AbstractBatchSystem(ABC):
         :param setOption: A function with signature setOption(varName, parsingFn=None, checkFn=None, default=None)
            used to update run configuration
         """
-        
+
     def getWorkerContexts(self):
         """
         Get a list of picklable context manager objects to wrap worker work in,
         in order.
-        
+
         Can be used to ask the Toil worker to do things in-process (such as
         configuring environment variables, hot-deploying user scripts, or
         cleaning up a node) that would otherwise require a wrapping "executor"
         process.
-        
+
         :rtype: list
         """
         return []
@@ -254,7 +253,7 @@ class BatchSystemSupport(AbstractBatchSystem):
     Partial implementation of AbstractBatchSystem, support methods.
     """
 
-    def __init__(self, config, maxCores, maxMemory, maxDisk):
+    def __init__(self, config: Config, maxCores: float, maxMemory: int, maxDisk: int):
         """
         Initializes initial state of the object
 
@@ -276,10 +275,7 @@ class BatchSystemSupport(AbstractBatchSystem):
         self.maxCores = maxCores
         self.maxMemory = maxMemory
         self.maxDisk = maxDisk
-        self.environment = {}
-        """
-        :type: dict[str,str]
-        """
+        self.environment: Dict[str, str] = {}
         self.workerCleanupInfo = WorkerCleanupInfo(workDir=self.config.workDir,
                                                    workflowID=self.config.workflowID,
                                                    cleanWorkDir=self.config.cleanWorkDir)
@@ -293,9 +289,9 @@ class BatchSystemSupport(AbstractBatchSystem):
         :param float cores: number of cores being requested
 
         :param int disk: amount of disk space being requested, in bytes
-        
+
         :param str job_name: Name of the job being checked, for generating a useful error report.
-        
+
         :param str detail: Batch-system-specific message to include in the error.
 
         :raise InsufficientSystemResources: raised when a resource is requested in an amount
@@ -347,10 +343,10 @@ class BatchSystemSupport(AbstractBatchSystem):
             try:
                 value = os.environ[name]
             except KeyError:
-                raise RuntimeError("%s does not exist in current environment", name)
+                raise RuntimeError(f"{name} does not exist in current environment")
         self.environment[name] = value
 
-    def formatStdOutErrPath(self, jobID, batchSystem, batchJobIDfmt, fileDesc):
+    def formatStdOutErrPath(self, toil_job_id: int, cluster_job_id: str, std: str) -> str:
         """
         Format path for batch system standard output/error and other files
         generated by the batch system itself.
@@ -359,28 +355,22 @@ class BatchSystemSupport(AbstractBatchSystem):
         be on a shared file system) with names containing both the Toil and
         batch system job IDs, for ease of debugging job failures.
 
-        :param: string jobID : Toil job ID
-        :param: string batchSystem : Name of the batch system
-        :param: string batchJobIDfmt : A string which the particular batch system
-            will format into the batch job ID once it is submitted
-        :param: string fileDesc : File description, should be 'std_output' for standard
-             output, 'std_error' for standard error, and as appropriate for other files
+        :param: int toil_job_id : The unique id that Toil gives a job.
+        :param: cluster_job_id : What the cluster, for example, GridEngine, uses as its internal job id.
+        :param: string std : The provenance of the stream (for example: 'err' for 'stderr' or 'out' for 'stdout')
 
         :rtype: string : Formatted filename; however if self.config.noStdOutErr is true,
              returns '/dev/null' or equivalent.
-
         """
         if self.config.noStdOutErr:
             return os.devnull
 
-        workflowID = self.config.workflowID
-        workDir = Toil.getToilWorkDir(self.config.workDir)
-        fileName = 'toil_workflow_{workflowID}_job_{jobID}_batch_{batchSystem}_{batchJobIDfmt}_{fileDesc}.log'.format(
-            workflowID=workflowID, jobID=jobID, batchSystem=batchSystem, batchJobIDfmt=batchJobIDfmt, fileDesc=fileDesc)
+        fileName: str = f'toil_{self.config.workflowID}.{toil_job_id}.{cluster_job_id}.{std}.log'
+        workDir: str = Toil.getToilWorkDir(self.config.workDir)
         return os.path.join(workDir, fileName)
 
     @staticmethod
-    def workerCleanup(info):
+    def workerCleanup(info: WorkerCleanupInfo) -> None:
         """
         Cleans up the worker node on batch system shutdown. Also see :meth:`supportsWorkerCleanup`.
 
@@ -455,56 +445,57 @@ class BatchSystemLocalSupport(BatchSystemSupport):
         """To be called from shutdown()"""
         self.localBatch.shutdown()
 
+
 class BatchSystemCleanupSupport(BatchSystemLocalSupport):
     """
     Adds cleanup support when the last running job leaves a node, for batch
     systems that can't provide it using the backing scheduler.
     """
-    
+
     @classmethod
     def supportsWorkerCleanup(cls):
         return True
-        
+
     def getWorkerContexts(self):
         # Tell worker to register for and invoke cleanup
-        
+
         # Create a context manager that has a copy of our cleanup info
         context = WorkerCleanupContext(self.workerCleanupInfo)
-       
+
         # Send it along so the worker works inside of it
         contexts = super(BatchSystemCleanupSupport, self).getWorkerContexts()
         contexts.append(context)
         return contexts
-        
+
     def __init__(self, config, maxCores, maxMemory, maxDisk):
         super(BatchSystemCleanupSupport, self).__init__(config, maxCores, maxMemory, maxDisk)
-        
+
 class WorkerCleanupContext:
     """
     Context manager used by :class:`BatchSystemCleanupSupport` to implement
     cleanup on a node after the last worker is done working.
-    
+
     Gets wrapped around the worker's work.
     """
-    
+
     def __init__(self, workerCleanupInfo):
         """
         Wrap the given workerCleanupInfo in a context manager.
-        
+
         :param WorkerCleanupInfo workerCleanupInfo: Info to use to clean up the worker if we are
                                                     the last to exit the context manager.
         """
-        
-        
+
+
         self.workerCleanupInfo = workerCleanupInfo
         self.arena = None
-        
+
     def __enter__(self):
         # Set up an arena so we know who is the last worker to leave
-        self.arena = LastProcessStandingArena(Toil.getToilWorkDir(self.workerCleanupInfo.workDir), 
+        self.arena = LastProcessStandingArena(Toil.getToilWorkDir(self.workerCleanupInfo.workDir),
                                               self.workerCleanupInfo.workflowID + '-cleanup')
         self.arena.enter()
-        
+
     def __exit__(self, type, value, traceback):
         for _ in self.arena.leave():
             # We are the last concurrent worker to finish.
