@@ -25,18 +25,22 @@ import re
 import subprocess
 from datetime import datetime
 from random import randint
+from typing import Union, Optional, List, Dict
 
 from dateutil.parser import parse
 from dateutil.tz import tzlocal
 
-from toil.batchSystems import MemoryString
 from toil.batchSystems.abstractBatchSystem import BatchJobExitReason
-from toil.batchSystems.abstractGridEngineBatchSystem import \
-    AbstractGridEngineBatchSystem
-from toil.batchSystems.lsfHelper import (check_lsf_json_output_supported,
-                                         parse_memory_limit,
-                                         parse_memory_resource,
-                                         per_core_reservation)
+from toil.batchSystems.abstractGridEngineBatchSystem import (
+    AbstractGridEngineBatchSystem,
+)
+from toil.batchSystems.lsfHelper import (
+    check_lsf_json_output_supported,
+    parse_mem_and_cmd_from_output,
+    parse_memory_limit,
+    parse_memory_resource,
+    per_core_reservation,
+)
 from toil.lib.misc import call_command
 
 logger = logging.getLogger(__name__)
@@ -81,12 +85,23 @@ class LSFBatchSystem(AbstractGridEngineBatchSystem):
         def killJob(self, jobID):
             call_command(['bkill', self.getBatchSystemID(jobID)])
 
-        def prepareSubmission(self, cpu, memory, jobID, command, jobName):
-            return self.prepareBsub(cpu, memory, jobID) + [command]
+        def prepareSubmission(self,
+                              cpu: int,
+                              memory: int,
+                              jobID: int,
+                              command: str,
+                              jobName: str,
+                              job_environment: Optional[Dict[str, str]] = None):
+            return (self.prepareBsub(cpu, memory, jobID) + [command],
+                    job_environment)  # pass job_environment to .submitJob()
 
         def submitJob(self, subLine):
+            subLine, job_environment = subLine
             combinedEnv = self.boss.environment
             combinedEnv.update(os.environ)
+            if job_environment:
+                combinedEnv.update(job_environment)
+
             stdout = call_command(subLine, env=combinedEnv)
             # Example success: Job <39605914> is submitted to default queue <general>.
             # Example fail: Service class does not exist. Job not submitted.
@@ -98,10 +113,57 @@ class LSFBatchSystem(AbstractGridEngineBatchSystem):
             else:
                 logger.error("Could not submit job\nReason: {}".format(stdout))
                 temp_id = randint(10000000, 99999999)
-                #Flag this job to be handled by getJobExitCode
+                # Flag this job to be handled by getJobExitCode
                 result = "NOT_SUBMITTED_{}".format(temp_id)
             return result
 
+        def coalesce_job_exit_codes(self, batch_job_id_list: list) -> list:
+            status_dict = {}
+            valid_batch_job_id_list = []
+            status_resonse = []
+            for single_lsf_id in batch_job_id_list:
+                if "NOT_SUBMITTED" in single_lsf_id:
+                    logger.error(
+                        "bjobs detected job [%s] failed to submit", single_lsf_id
+                    )
+                    status_dict[single_lsf_id] = 1
+                job = single_lsf_id
+                if "." in single_lsf_id:
+                    job = single_lsf_id.split(".", 1)[0]
+                valid_batch_job_id_list.append(job)
+            if valid_batch_job_id_list:
+                args = [
+                    "bjobs",
+                    "-json",
+                    "-o",
+                    "jobid user exit_code stat exit_reason pend_reason",
+                ] + valid_batch_job_id_list
+                logger.debug("Getting coalesced job exit codes via bjobs")
+                bjobs_records = self.parseBjobs(
+                    subprocess.run(
+                        args, check=False, stderr=subprocess.STDOUT, encoding="utf-8"
+                    ).stdout
+                )
+                if bjobs_records:
+                    for single_record in bjobs_records:
+                        if "JOBID" in single_record:
+                            single_job_id = single_record["JOBID"]
+                            status_dict[single_job_id] = self.parse_bjobs_record(
+                                single_record, single_job_id
+                            )
+            for single_lsf_id in batch_job_id_list:
+                if "NOT_SUBMITTED" in single_lsf_id:
+                    status_resonse.append(status_dict[single_lsf_id])
+                else:
+                    job = single_lsf_id
+                    if "." in single_lsf_id:
+                        job = single_lsf_id.split(".", 1)[0]
+                    if job in status_dict:
+                        status_resonse.append(status_dict[job])
+                    else:
+                        status_resonse.append(None)
+            return status_resonse
+
         def getJobExitCode(self, lsfJobID):
             # the task is set as part of the job ID if using getBatchSystemID()
             if "NOT_SUBMITTED" in lsfJobID:
@@ -123,52 +185,60 @@ class LSFBatchSystem(AbstractGridEngineBatchSystem):
                 bjobs_records = self.parseBjobs(stdout)
                 if bjobs_records:
                     process_output = bjobs_records[0]
-                    if 'STAT' in process_output:
-                        process_status = process_output['STAT']
-                        if process_status == 'DONE':
-                            logger.debug(
-                                "bjobs detected job completed for job: {}".format(job))
-                            return 0
-                        if process_status == 'PEND':
-                            pending_info = ""
-                            if 'PEND_REASON' in process_output:
-                                if process_output['PEND_REASON']:
-                                    pending_info = "\n" + \
-                                        process_output['PEND_REASON']
-                            logger.debug(
-                                "bjobs detected job pending with: {}\nfor job: {}".format(pending_info, job))
-                            return None
-                        if process_status == 'EXIT':
-                            exit_code = 1
-                            exit_reason = ""
-                            if 'EXIT_CODE' in process_output:
-                                exit_code_str = process_output['EXIT_CODE']
-                                if exit_code_str:
-                                    exit_code = int(exit_code_str)
-                            if 'EXIT_REASON' in process_output:
-                                exit_reason = process_output['EXIT_REASON']
-                            exit_info = ""
-                            if exit_code:
-                                exit_info = "\nexit code: {}".format(exit_code)
-                            if exit_reason:
-                                exit_info += "\nexit reason: {}".format(exit_reason)
-                            logger.error(
-                                "bjobs detected job failed with: {}\nfor job: {}".format(exit_info, job))
-                            if "TERM_MEMLIMIT" in exit_reason:
-                                return BatchJobExitReason.MEMLIMIT
-                            return exit_code
-                        if process_status == 'RUN':
-                            logger.debug(
-                                "bjobs detected job started but not completed for job: {}".format(job))
-                            return None
-                        if process_status in {'PSUSP', 'USUSP', 'SSUSP'}:
-                            logger.debug(
-                                "bjobs detected job suspended for job: {}".format(job))
-                            return None
+                    return self.parse_bjobs_record(process_output, job)
 
-                        return self.getJobExitCodeBACCT(job)
-            else:
-                return self.fallbackGetJobExitCode(job)
+            return self.fallbackGetJobExitCode(job)
+
+        def parse_bjobs_record(self, bjobs_record: dict, job: int) -> Union[int, None]:
+            """
+            Helper functions for getJobExitCode and  to parse the bjobs status record
+            """
+            if "STAT" in bjobs_record:
+                process_status = bjobs_record["STAT"]
+                if process_status == "DONE":
+                    logger.debug("bjobs detected job completed for job: %s", job)
+                    return 0
+                if process_status == "PEND":
+                    pending_info = ""
+                    if "PEND_REASON" in bjobs_record:
+                        if bjobs_record["PEND_REASON"]:
+                            pending_info = "\n" + bjobs_record["PEND_REASON"]
+                    logger.debug(
+                        "bjobs detected job pending with: %s\nfor job: %s",
+                        (pending_info, job),
+                    )
+                    return None
+                if process_status == "EXIT":
+                    exit_code = 1
+                    exit_reason = ""
+                    if "EXIT_CODE" in bjobs_record:
+                        exit_code_str = bjobs_record["EXIT_CODE"]
+                        if exit_code_str:
+                            exit_code = int(exit_code_str)
+                    if "EXIT_REASON" in bjobs_record:
+                        exit_reason = bjobs_record["EXIT_REASON"]
+                    exit_info = ""
+                    if exit_code:
+                        exit_info = "\nexit code: {}".format(exit_code)
+                    if exit_reason:
+                        exit_info += "\nexit reason: {}".format(exit_reason)
+                    logger.error(
+                        "bjobs detected job failed with: %s\nfor job: %s",
+                        (exit_info, job),
+                    )
+                    if "TERM_MEMLIMIT" in exit_reason:
+                        return BatchJobExitReason.MEMLIMIT
+                    return exit_code
+                if process_status == "RUN":
+                    logger.debug(
+                        "bjobs detected job started but not completed for job: %s", job
+                    )
+                    return None
+                if process_status in {"PSUSP", "USUSP", "SSUSP"}:
+                    logger.debug("bjobs detected job suspended for job: %s", job)
+                    return None
+
+                return self.getJobExitCodeBACCT(job)
 
         def getJobExitCodeBACCT(self,job):
             # if not found in bjobs, then try bacct (slower than bjobs)
@@ -193,40 +263,32 @@ class LSFBatchSystem(AbstractGridEngineBatchSystem):
 
         def fallbackGetJobExitCode(self, job):
             args = ["bjobs", "-l", str(job)]
-            logger.debug("Checking job exit code for job via bjobs (fallback): "
-                         "{}".format(job))
+            logger.debug(f"Checking job exit code for job via bjobs (fallback): {job}")
             stdout = call_command(args)
             output = stdout.replace("\n                     ", "")
             process_output = output.split('\n')
             started = 0
             for line in process_output:
                 if "Done successfully" in line or "Status <DONE>" in line:
-                    logger.debug("bjobs detected job completed for job: "
-                                 "{}".format(job))
+                    logger.debug(f"bjobs detected job completed for job: {job}")
                     return 0
                 elif "New job is waiting for scheduling" in line:
-                    logger.debug("bjobs detected job pending scheduling for "
-                                 "job: {}".format(job))
+                    logger.debug(f"bjobs detected job pending scheduling for job: {job}")
                     return None
                 elif "PENDING REASONS" in line or "Status <PEND>" in line:
-                    logger.debug("bjobs detected job pending for job: "
-                                 "{}".format(job))
+                    logger.debug(f"bjobs detected job pending for job: {job}")
                     return None
                 elif "Exited with exit code" in line:
-                    exit = int(line[line.find("Exited with exit code ")+22:]
-                               .split('.')[0])
-                    logger.error("bjobs detected job exit code "
-                                 "{} for job {}".format(exit, job))
+                    exit = int(line[line.find("Exited with exit code ")+22:].split('.')[0])
+                    logger.error(f"bjobs detected job exit code {exit} for job {job}")
                     return exit
                 elif "Completed <exit>" in line:
-                    logger.error("bjobs detected job failed for job: "
-                                 "{}".format(job))
+                    logger.error(f"bjobs detected job failed for job: {job}")
                     return 1
                 elif line.find("Started on ") > -1 or "Status <RUN>" in line:
                     started = 1
             if started == 1:
-                logger.debug("bjobs detected job started but not completed: "
-                             "{}".format(job))
+                logger.debug(f"bjobs detected job started but not completed: {job}")
                 return None
 
             return self.getJobExitCodeBACCT(job)
@@ -234,7 +296,7 @@ class LSFBatchSystem(AbstractGridEngineBatchSystem):
         """
         Implementation-specific helper methods
         """
-        def prepareBsub(self, cpu, mem, jobID):
+        def prepareBsub(self, cpu: int, mem: int, jobID: int) -> List[str]:
             """
             Make a bsub commandline to execute.
 
@@ -243,34 +305,30 @@ class LSFBatchSystem(AbstractGridEngineBatchSystem):
               mem: number of bytes of memory needed
               jobID: ID number of the job
             """
+            bsubMem = []
             if mem:
-                if per_core_reservation():
-                    mem = float(mem)/1024**3/math.ceil(cpu)
-                    mem_resource = parse_memory_resource(mem)
-                    mem_limit = parse_memory_limit(mem)
-                else:
-                    mem = float(mem) / 1024**3
-                    mem_resource = parse_memory_resource(mem)
-                    mem_limit = parse_memory_limit(mem)
-
-                bsubMem = ['-R', 'select[mem > {m}] '
-                           'rusage[mem={m}]'.format(m=mem_resource),
-                           '-M', str(mem_limit)]
-            else:
-                bsubMem = []
+                mem = float(mem) / 1024 ** 3
+                if per_core_reservation() and cpu:
+                    mem = mem / math.ceil(cpu)
+                mem_resource = parse_memory_resource(mem)
+                mem_limit = parse_memory_limit(mem)
+                bsubMem = ['-R',
+                           f'select[mem>{mem_resource}] '
+                           f'rusage[mem={mem_resource}]',
+                           '-M', mem_limit]
             bsubCpu = [] if cpu is None else ['-n', str(math.ceil(cpu))]
-            bsubline = ["bsub", "-cwd", ".", "-J", "toil_job_{}".format(jobID)]
+            bsubline = ["bsub", "-cwd", ".", "-J", f"toil_job_{jobID}"]
             bsubline.extend(bsubMem)
             bsubline.extend(bsubCpu)
-            stdoutfile = self.boss.formatStdOutErrPath(jobID, 'lsf', '%J', 'std_output')
-            stderrfile = self.boss.formatStdOutErrPath(jobID, 'lsf', '%J', 'std_error')
+            stdoutfile: str = self.boss.formatStdOutErrPath(jobID, '%J', 'out')
+            stderrfile: str = self.boss.formatStdOutErrPath(jobID, '%J', 'err')
             bsubline.extend(['-o', stdoutfile, '-e', stderrfile])
             lsfArgs = os.getenv('TOIL_LSF_ARGS')
             if lsfArgs:
                 bsubline.extend(lsfArgs.split())
             return bsubline
 
-        def parseBjobs(self,bjobs_output_str):
+        def parseBjobs(self, bjobs_output_str):
             """
             Parse records from bjobs json type output
             params:
@@ -298,67 +356,23 @@ class LSFBatchSystem(AbstractGridEngineBatchSystem):
             """
             Parse the maximum memory from job.
 
-            params:
-            jobID: ID number of the job
+            :param jobID: ID number of the job
             """
-            memargs = ["bjobs", "-l", str(jobID)]
             try:
-                bjobs = subprocess.check_output(memargs, universal_newlines=True)
-                memregex = r"MAX MEM: (.*?);"
-                meminfo = re.search(memregex, bjobs)
-                s = " ".join(bjobs.split())
-                command = re.search(r"Command <(.*?)>", s)
-                if meminfo:
-                    if not command:
-                        logger.info("Cannot Parse Max Memory Due to Missing Command String: %s", bjobs)
-                    else:
-                        logger.info("[job ID %s, Command %s] the maximum memory used was: %s",
-                                    str(jobID), command.group(1), meminfo.group(1))
+                output = subprocess.check_output(["bjobs", "-l", str(jobID)], universal_newlines=True)
+                max_mem, command = parse_mem_and_cmd_from_output(output=output)
+                if not max_mem:
+                    logger.warning(f"[job ID {jobID}] Unable to Collect Maximum Memory Usage: {output}")
+                    return
+
+                if not command:
+                    logger.warning(f"[job ID {jobID}] Cannot Parse Max Memory Due to Missing Command String: {output}")
                 else:
-                    logger.debug("[job ID %s] Unable to collect maximum memory usage: %s",
-                                 str(jobID), bjobs)
-                return meminfo
-            except subprocess.CalledProcessError as err:
-                logger.debug("[job ID %s] Unable to collect maximum memory usage: %s",
-                             str(jobID), str(err))
+                    logger.info(f"[job ID {jobID}, Command {command.group(1)}] Max Memory Used: {max_mem.group(1)}")
+                return max_mem
+            except subprocess.CalledProcessError as e:
+                logger.warning(f"[job ID {jobID}] Unable to Collect Maximum Memory Usage: {e}")
 
     def getWaitDuration(self):
         """We give LSF a second to catch its breath (in seconds)"""
         return 60
-
-    @classmethod
-    def obtainSystemConstants(cls):
-        stdout = call_command(["lshosts"])
-        line = stdout.split('\n')[0]
-        items = line.strip().split()
-        num_columns = len(items)
-        cpu_index = None
-        mem_index = None
-        for i in range(num_columns):
-            if items[i] == 'ncpus':
-                cpu_index = i
-            elif items[i] == 'maxmem':
-                mem_index = i
-
-        if cpu_index is None or mem_index is None:
-            raise RuntimeError("lshosts command does not return ncpus or maxmem columns")
-
-        maxCPU = 0
-        maxMEM = MemoryString("0")
-        for line in stdout.split('\n')[1:]:
-            items = line.strip().split()
-            if items:
-                if len(items) < num_columns:
-                    raise RuntimeError("lshosts output has a varying number of columns")
-                if items[cpu_index] != '-' and int(items[cpu_index]) > int(maxCPU):
-                    maxCPU = int(items[cpu_index])
-                if items[mem_index] != '-' and MemoryString(items[mem_index]) > maxMEM:
-                    maxMEM = MemoryString(items[mem_index])
-
-        if maxCPU == 0 or maxMEM == MemoryString("0"):
-                raise RuntimeError("lshosts returns null ncpus or maxmem info")
-
-        logger.debug("Got the maxMEM: {}".format(maxMEM))
-        logger.debug("Got the maxCPU: {}".format(maxCPU))
-
-        return maxCPU, maxMEM
