@@ -14,6 +14,7 @@
 import logging
 import math
 import os
+import signal
 import subprocess
 import sys
 import time
@@ -21,8 +22,10 @@ import traceback
 from contextlib import contextmanager
 from queue import Empty, Queue
 from threading import Condition, Event, Lock, Thread
+from typing import Dict, List, Optional, Sequence
 
 import toil
+import toil.job
 from toil import worker as toil_worker
 from toil.batchSystems.abstractBatchSystem import (EXIT_STATUS_UNAVAILABLE_VALUE,
                                                    BatchSystemSupport,
@@ -65,14 +68,14 @@ class SingleMachineBatchSystem(BatchSystemSupport):
     minCores = 0.1
     """
     The minimal fractional CPU. Tasks with a smaller core requirement will be rounded up to this
-    value. 
+    value.
     """
     physicalMemory = toil.physicalMemory()
 
     def __init__(self, config, maxCores, maxMemory, maxDisk):
         self.config = config
         # Limit to the smaller of the user-imposed limit and what we actually
-        # have on this machine for each resource. 
+        # have on this machine for each resource.
         #
         # If we don't have up to the limit of the resource (and the resource
         # isn't the inlimited sentinel), warn.
@@ -94,7 +97,7 @@ class SingleMachineBatchSystem(BatchSystemSupport):
                 # We have an actually specified limit and not the default
                 log.warning('Not enough disk space! User limited to %i bytes but we only have %i bytes.', maxDisk, self.physicalDisk)
             maxDisk = self.physicalDisk
-            
+
         super(SingleMachineBatchSystem, self).__init__(config, maxCores, maxMemory, maxDisk)
         assert self.maxCores >= self.minCores
         assert self.maxMemory >= 1
@@ -103,24 +106,21 @@ class SingleMachineBatchSystem(BatchSystemSupport):
         # squeezing more tasks onto each core (scale < 1) or stretching tasks over more cores
         # (scale > 1).
         self.scale = config.scale
-        
+
         if config.badWorker > 0 and config.debugWorker:
             # We can't throw SIGUSR1 at the worker because it is also going to
             # be the leader and/or test harness.
             raise RuntimeError("Cannot use badWorker and debugWorker together; "
                 "worker would have to kill the leader")
-                
+
         self.debugWorker = config.debugWorker
-        
+
         # A counter to generate job IDs and a lock to guard it
         self.jobIndex = 0
         self.jobIndexLock = Lock()
 
         # A dictionary mapping IDs of submitted jobs to the command line
-        self.jobs = {}
-        """
-        :type: dict[str,toil.job.JobDescription]
-        """
+        self.jobs: Dict[str, toil.job.JobDescription] = {}
 
         # A queue of jobs waiting to be executed. Consumed by the daddy thread.
         self.inputQueue = Queue()
@@ -129,24 +129,15 @@ class SingleMachineBatchSystem(BatchSystemSupport):
         self.outputQueue = Queue()
 
         # A dictionary mapping IDs of currently running jobs to their Info objects
-        self.runningJobs = {}
-        """
-        :type: dict[str,Info]
-        """
-        
+        self.runningJobs: Dict[str, Info] = {}
+
         # These next two are only used outside debug-worker mode
 
         # A dict mapping PIDs to Popen objects for running jobs.
         # Jobs that don't fork are executed one at a time in the main thread.
-        self.children = {}
-        """
-        :type: dict[int,subprocess.Popen]
-        """
+        self.children: Dict[int, subprocess.Popen] = {}
         # A dict mapping child PIDs to the Job IDs they are supposed to be running.
-        self.childToJob = {}
-        """
-        :type: dict[int,str]
-        """
+        self.childToJob: Dict[int, str] = {}
 
         # A pool representing available CPU in units of minCores
         self.coreFractions = ResourcePool(int(self.maxCores / self.minCores), 'cores')
@@ -154,7 +145,7 @@ class SingleMachineBatchSystem(BatchSystemSupport):
         self.memory = ResourcePool(self.maxMemory, 'memory')
         # A pool representing the available space in bytes
         self.disk = ResourcePool(self.maxDisk, 'disk')
-        
+
         # If we can't schedule something, we fill this in with a reason why
         self.schedulingStatusMessage = None
 
@@ -165,21 +156,21 @@ class SingleMachineBatchSystem(BatchSystemSupport):
         # Also takes care of resource accounting.
         self.daddyThread = None
         # If it breaks it will fill this in
-        self.daddyException = None
+        self.daddyException: Optional[Exception] = None
 
         if self.debugWorker:
-            log.debug('Started in worker debug mode.')
+            log.debug('Started batch system %s in worker debug mode.', id(self))
         else:
             self.daddyThread = Thread(target=self.daddy, daemon=True)
             self.daddyThread.start()
-            log.debug('Started in normal mode.')
-            
+            log.debug('Started batch system %s in normal mode.', id(self))
+
     def daddy(self):
         """
         Be the "daddy" thread.
 
         Our job is to look at jobs from the input queue.
-        
+
         If a job fits in the available resources, we allocate resources for it
         and kick off a child process.
 
@@ -190,7 +181,7 @@ class SingleMachineBatchSystem(BatchSystemSupport):
         """
 
         try:
-            log.debug('Started daddy thread.')
+            log.debug('Started daddy thread for batch system %s.', id(self))
 
             while not self.shuttingDown.is_set():
                 # Main loop
@@ -203,7 +194,7 @@ class SingleMachineBatchSystem(BatchSystemSupport):
                         jobCommand, jobID, jobCores, jobMemory, jobDisk, environment = args
 
                         coreFractions = int(jobCores / self.minCores)
-                        
+
                         # Try to start the child
                         result = self._startChild(jobCommand, jobID,
                             coreFractions, jobMemory, jobDisk, environment)
@@ -234,21 +225,19 @@ class SingleMachineBatchSystem(BatchSystemSupport):
                 # TODO: It would be good to be able to wait on a new job or a finished child, whichever comes first.
                 # For now we just sleep and loop.
                 time.sleep(0.01)
-                    
+
 
             # When we get here, we are shutting down.
-            
-            for popen in self.children.values():
-                # Kill all the children, going through popen to avoid signaling re-used PIDs.
-                popen.kill()
-            for popen in self.children.values():
-                # Reap all the children
-                popen.wait()
-            
+            log.debug('Daddy thread cleaning up %d remaining children for batch system %s...', len(self.children), id(self))
+
+            self._stop_and_wait(self.children.values())
+
+            log.debug('Daddy thread for batch system %s finishing because no children should now exist', id(self))
+
             # Then exit the thread.
             return
         except Exception as e:
-            log.critical('Unhandled exception in daddy thread: %s', traceback.format_exc())
+            log.critical('Unhandled exception in daddy thread for batch system %s: %s', id(self), traceback.format_exc())
             # Pass the exception back to the main thread so it can stop the next person who calls into us.
             self.daddyException = e
             raise
@@ -259,15 +248,98 @@ class SingleMachineBatchSystem(BatchSystemSupport):
             log.critical('Propagating unhandled exception in daddy thread to main thread')
             exc = self.daddyException
             self.daddyException = None
-            raise exc
+            if isinstance(exc, Exception):
+                raise exc
+            else:
+                raise TypeError(f'Daddy thread failed with non-exception: {exc}')
+
+    def _stop_now(self, popens: Sequence[subprocess.Popen]) -> List[int]:
+        """
+        Stop the given child processes and all their children. Does not reap them.
+
+        Returns a list of PGIDs killed, where processes may exist that have not
+        yet received their kill signals.
+        """
+
+        # We will potentially need to poll these PGIDs to ensure that all
+        # processes in them are gone.
+        pgids = []
+
+        for popen in popens:
+            # Kill all the children
+
+            if popen.returncode is None:
+                # Process is not known to be dead. Try and grab its group.
+                try:
+                    pgid = os.getpgid(popen.pid)
+                except OSError:
+                    # It just died. Assume the pgid was its PID.
+                    pgid = popen.pid
+            else:
+                # It is dead. Try it's PID as a PGID and hope we didn't re-use it.
+                pgid = popen.pid
+
+            if pgid != os.getpgrp():
+                # The child process really is in its own group, and not ours.
+
+                # Kill the group, which hopefully hasn't been reused
+                log.debug('Send shutdown kill to process group %s known to batch system %s', pgid, id(self))
+                try:
+                    os.killpg(pgid, signal.SIGKILL)
+                    pgids.append(pgid)
+                except ProcessLookupError:
+                    # It is dead already
+                    pass
+                except PermissionError:
+                    # It isn't ours actually. Ours is dead.
+                    pass
+            else:
+                # Kill the subprocess again through popen in case it somehow
+                # never managed to make the group.
+                popen.kill()
+
+        return pgids
+
+    def _stop_and_wait(self, popens: Sequence[subprocess.Popen]) -> None:
+        """
+        Stop the given child processes and all their children. Blocks until the
+        processes are gone.
+        """
+
+        pgids = self._stop_now(popens)
+
+        for popen in popens:
+            # Wait on all the children
+            popen.wait()
+
+            log.debug('Process %s known to batch system %s is stopped; it returned %s', popen.pid, id(self), popen.returncode)
+
+        for pgid in pgids:
+            try:
+                while True:
+                    # Send a kill to the group again, to see if anything in it
+                    # is still alive. Our first kill might not have been
+                    # delivered yet.
+                    os.killpg(pgid, signal.SIGKILL)
+                    # If that worked it is still alive, so wait for the kernel
+                    # to stop fooling around and kill it.
+                    log.warning('Sent redundant shutdown kill to surviving process group %s known to batch system %s', pgid, id(self))
+                    time.sleep(0.1)
+            except ProcessLookupError:
+                # The group is actually gone now.
+                pass
+            except PermissionError:
+                # The group is not only gone but reused
+                pass
+
 
     def _pollForDoneChildrenIn(self, pid_to_popen):
         """
         See if any children represented in the given dict from PID to Popen
         object have finished.
-        
+
         Return a collection of their PIDs.
-        
+
         Guarantees that each child's exit code will be gettable via wait() on
         the child's Popen object (i.e. does not reap the child, unless via
         Popen).
@@ -276,7 +348,7 @@ class SingleMachineBatchSystem(BatchSystemSupport):
         # We keep our found PIDs in a set so we can work around waitid showing
         # us the same one repeatedly.
         ready = set()
-        
+
         # Find the waitid function
         waitid = getattr(os, 'waitid', None)
 
@@ -313,7 +385,7 @@ class SingleMachineBatchSystem(BatchSystemSupport):
                     # Process is done
                     ready.add(pid)
                     log.debug('Child %d has stopped', pid)
-            
+
             # Return all the done processes we found
             return ready
 
@@ -323,10 +395,7 @@ class SingleMachineBatchSystem(BatchSystemSupport):
         May only be called in debug-worker mode.
         Assumes resources are available.
         """
-        
         assert self.debugWorker
-
-    
         # TODO: It is not possible to kill running jobs in forkless mode,
         # because they are run immediately in the main thread.
         info = Info(time.time(), None, None, killIntended=False)
@@ -334,9 +403,9 @@ class SingleMachineBatchSystem(BatchSystemSupport):
 
         if jobCommand.startswith("_toil_worker "):
             # We can actually run in this thread
-            jobName, jobStoreLocator, jobStoreID = jobCommand.split()[1:] # Parse command
+            jobName, jobStoreLocator, jobStoreID = jobCommand.split()[1:4] # Parse command
             jobStore = Toil.resumeJobStore(jobStoreLocator)
-            toil_worker.workerScript(jobStore, jobStore.config, jobName, jobStoreID, 
+            toil_worker.workerScript(jobStore, jobStore.config, jobName, jobStoreID,
                                      redirectOutputToLogFile=not self.debugWorker) # Call the worker
         else:
             # Run synchronously. If starting or running the command fails, let the exception stop us.
@@ -347,24 +416,24 @@ class SingleMachineBatchSystem(BatchSystemSupport):
         self.runningJobs.pop(jobID)
         if not info.killIntended:
             self.outputQueue.put(UpdatedBatchJobInfo(jobID=jobID, exitStatus=0, wallTime=time.time() - info.time, exitReason=None))
-            
+
     def getSchedulingStatusMessage(self):
         # Implement the abstractBatchSystem's scheduling status message API
         return self.schedulingStatusMessage
-    
+
     def _setSchedulingStatusMessage(self, message):
         """
         If we can't run a job, we record a short message about why not. If the
         leader wants to know what is up with us (for example, to diagnose a
         deadlock), it can ask us for the message.
         """
-        
+
         self.schedulingStatusMessage = message
 
     def _startChild(self, jobCommand, jobID, coreFractions, jobMemory, jobDisk, environment):
         """
         Start a child process for the given job.
-        
+
         Allocate its required resources and save it and save it in our bookkeeping structures.
 
         If the job is started, returns its PID.
@@ -390,10 +459,18 @@ class SingleMachineBatchSystem(BatchSystemSupport):
                     # So it is important to not lose track of the child process.
 
                     try:
-                        # Launch the job
+                        # Launch the job.
+                        # Make sure it is in its own session (and thus its own
+                        # process group) so that, if the user signals the
+                        # workflow, Toil will be responsible for killing the
+                        # job. This also makes sure that we can signal the job
+                        # and all its children together. We assume that the
+                        # process group ID will equal the PID of the process we
+                        # are starting.
                         popen = subprocess.Popen(jobCommand,
                                                  shell=True,
-                                                 env=dict(os.environ, **environment))
+                                                 env=dict(os.environ, **environment),
+                                                 start_new_session=True)
                     except Exception:
                         # If the job can't start, make sure we release resources now
                         self.coreFractions.release(coreFractions)
@@ -401,7 +478,7 @@ class SingleMachineBatchSystem(BatchSystemSupport):
                         self.disk.release(jobDisk)
 
                         log.error('Could not start job %s: %s', jobID, traceback.format_exc())
-                        
+
                         # Report as failed.
                         self.outputQueue.put(UpdatedBatchJobInfo(jobID=jobID, exitStatus=EXIT_STATUS_UNAVAILABLE_VALUE, wallTime=0, exitReason=None))
 
@@ -444,7 +521,7 @@ class SingleMachineBatchSystem(BatchSystemSupport):
         # Report that.
         return None
 
-    def _handleChild(self, pid):
+    def _handleChild(self, pid: int) -> None:
         """
         Handle a child process PID that has finished.
         The PID must be for a child job we started.
@@ -452,7 +529,7 @@ class SingleMachineBatchSystem(BatchSystemSupport):
 
         Remove the child from our bookkeeping structures and free its resources.
         """
-        
+
         # Look up the child
         popen = self.children[pid]
         jobID = self.childToJob[pid]
@@ -465,7 +542,22 @@ class SingleMachineBatchSystem(BatchSystemSupport):
         self.runningJobs.pop(jobID)
         self.childToJob.pop(pid)
         self.children.pop(pid)
-        
+
+        if popen.returncode is None or not callable(getattr(os, 'waitid', None)):
+            # It isn't reaped yet, or we have to reap all children to see if thay're done.
+            # Before we reap it (if possible), kill its PID as a PGID to make sure
+            # it isn't leaving children behind.
+            # TODO: This is a PGID re-use risk on Mac because the process is
+            # reaped already and the PGID may have been reused.
+            try:
+                os.killpg(pid, signal.SIGKILL)
+            except ProcessLookupError:
+                # It is dead already
+                pass
+            except PermissionError:
+                # It isn't ours actually. Ours is dead.
+                pass
+
         # See how the child did, and reap it.
         statusCode = popen.wait()
         if statusCode != 0 and not info.killIntended:
@@ -476,6 +568,31 @@ class SingleMachineBatchSystem(BatchSystemSupport):
             # If we killed it then it shouldn't show up in the queue.
             self.outputQueue.put(UpdatedBatchJobInfo(jobID=jobID, exitStatus=statusCode, wallTime=time.time() - info.time, exitReason=None))
 
+        # Make absolutely sure all processes in the group have received their
+        # kill signals and been cleaned up.
+        # TODO: this opens a PGID reuse risk; we reaped the process and its
+        # PGID may have been re-used. But it probably hasn't been and we
+        # definitely want to make sure all its children died before saying the
+        # job is done. Some might not be dead yet if we don't do this.
+        # TODO: can we safely do this before reaping? Or would we sit forever
+        # signaling a dead but unreaped process?
+        try:
+            while True:
+                # Send a kill to the group again, to see if anything in it
+                # is still alive. Our first kill might not have been
+                # delivered yet.
+                os.killpg(pid, signal.SIGKILL)
+                # If that worked it is still alive, so wait for the kernel
+                # to stop fooling around and kill it.
+                log.warning('Sent redundant job completion kill to surviving process group %s known to batch system %s', pid, id(self))
+                time.sleep(0.1)
+        except ProcessLookupError:
+            # It is dead already
+            pass
+        except PermissionError:
+            # It isn't ours actually. Ours is dead.
+            pass
+
         # Free up the job's resources.
         self.coreFractions.release(coreFractions)
         self.memory.release(jobMemory)
@@ -483,16 +600,15 @@ class SingleMachineBatchSystem(BatchSystemSupport):
 
         log.debug('Child %d for job %s succeeded', pid, jobID)
 
-        
-    def issueBatchJob(self, jobDesc):
+    def issueBatchJob(self, jobDesc, job_environment: Optional[Dict[str, str]] = None):
         """Adds the command and resources to a queue to be run."""
 
         self._checkOnDaddy()
 
         # Round cores to minCores and apply scale.
-        # Make sure to give minCores even if asked for 0 cores, or negative or something. 
+        # Make sure to give minCores even if asked for 0 cores, or negative or something.
         cores = max(math.ceil(jobDesc.cores * self.scale / self.minCores) * self.minCores, self.minCores)
-        
+
         # Don't do our own assertions about job size vs. our configured size.
         # The abstract batch system can handle it.
         self.checkResourceRequest(jobDesc.memory, cores, jobDesc.disk, job_name=jobDesc.jobName,
@@ -503,51 +619,63 @@ class SingleMachineBatchSystem(BatchSystemSupport):
             jobID = self.jobIndex
             self.jobIndex += 1
         self.jobs[jobID] = jobDesc.command
-        
+
+        environment = self.environment.copy()
+        if job_environment:
+            environment.update(job_environment)
+
         if self.debugWorker:
             # Run immediately, blocking for return.
             # Ignore resource requirements; we run one job at a time
-            self._runDebugJob(jobDesc.command, jobID, self.environment.copy())
+            self._runDebugJob(jobDesc.command, jobID, environment)
         else:
             # Queue the job for later
             self.inputQueue.put((jobDesc.command, jobID, cores, jobDesc.memory,
-                                jobDesc.disk, self.environment.copy()))
-            
+                                jobDesc.disk, environment))
 
         return jobID
 
-    def killBatchJobs(self, jobIDs):
+    def killBatchJobs(self, jobIDs: Sequence[str]) -> None:
         """Kills jobs by ID."""
 
         self._checkOnDaddy()
 
         log.debug('Killing jobs: {}'.format(jobIDs))
+
+        # Collect the popen handles for the jobs we have to stop
+        popens: List[subprocess.Popen] = []
+
         for jobID in jobIDs:
             if jobID in self.runningJobs:
                 info = self.runningJobs[jobID]
                 info.killIntended = True
                 if info.popen is not None:
-                    log.debug('Send kill to PID %s', info.popen.pid)
-                    info.popen.kill()
-                    log.debug('Sent kill to PID %s', info.popen.pid)
+                    popens.append(info.popen)
                 else:
-                    # No popen if running in forkless mode currently 
+                    # No popen if running in forkless mode currently
                     assert self.debugWorker
                     log.critical("Can't kill job: %s in debug mode" % jobID)
-                while jobID in self.runningJobs:
-                    pass
+
+        # Stop them all in a batch. Don't reap, because we need the daddy
+        # thread to reap them to mark the jobs as not running anymore.
+        self._stop_now(popens)
+
+        for jobID in jobIDs:
+            while jobID in self.runningJobs:
+                # Wait for the daddy thread to collect them.
+                time.sleep(0.01)
 
     def getIssuedBatchJobIDs(self):
         """Just returns all the jobs that have been run, but not yet returned as updated."""
-        
+
         self._checkOnDaddy()
-        
+
         return list(self.jobs.keys())
 
     def getRunningBatchJobIDs(self):
-        
+
         self._checkOnDaddy()
-        
+
         now = time.time()
         return {jobID: now - info.time for jobID, info in list(self.runningJobs.items())}
 
@@ -555,7 +683,7 @@ class SingleMachineBatchSystem(BatchSystemSupport):
         """
         Cleanly terminate and join daddy thread.
         """
-        
+
         if self.daddyThread is not None:
             # Tell the daddy thread to stop.
             self.shuttingDown.set()
@@ -566,9 +694,9 @@ class SingleMachineBatchSystem(BatchSystemSupport):
 
     def getUpdatedBatchJob(self, maxWait):
         """Returns a tuple of a no-longer-running job, the return value of its process, and its runtime, or None."""
-        
+
         self._checkOnDaddy()
-        
+
         try:
             item = self.outputQueue.get(timeout=maxWait)
         except Empty:
@@ -630,7 +758,7 @@ class ResourcePool:
             self.value -= amount
             self.__validate()
             return True
-            
+
 
     def acquire(self, amount):
         """
