@@ -15,13 +15,14 @@ import logging
 import math
 import os
 from pipes import quote
+from typing import List, Dict, Optional
 
-from toil.batchSystems import MemoryString
 from toil.batchSystems.abstractGridEngineBatchSystem import AbstractGridEngineBatchSystem
 from toil.lib.misc import CalledProcessErrorStderr, call_command
 
 logger = logging.getLogger(__name__)
 
+
 class SlurmBatchSystem(AbstractGridEngineBatchSystem):
 
     class Worker(AbstractGridEngineBatchSystem.Worker):
@@ -51,8 +52,14 @@ class SlurmBatchSystem(AbstractGridEngineBatchSystem):
         def killJob(self, jobID):
             call_command(['scancel', self.getBatchSystemID(jobID)])
 
-        def prepareSubmission(self, cpu, memory, jobID, command, jobName):
-            return self.prepareSbatch(cpu, memory, jobID, jobName) + ['--wrap={}'.format(command)]
+        def prepareSubmission(self,
+                              cpu: int,
+                              memory: int,
+                              jobID: int,
+                              command: str,
+                              jobName: str,
+                              job_environment: Optional[Dict[str, str]] = None) -> List[str]:
+            return self.prepareSbatch(cpu, memory, jobID, jobName, job_environment) + ['--wrap={}'.format(command)]
 
         def submitJob(self, subLine):
             try:
@@ -116,26 +123,26 @@ class SlurmBatchSystem(AbstractGridEngineBatchSystem):
 
             stdout = call_command(args)
             if isinstance(stdout, str):
-                values = stdout.strip().split()
+                lines = stdout.splitlines()
             elif isinstance(stdout, bytes):
-                values = stdout.decode('utf-8').strip().split()
-
-            # If job information is not available an error is issued:
-            # slurm_load_jobs error: Invalid job id specified
-            # There is no job information, so exit.
-            if len(values) > 0 and values[0] == 'slurm_load_jobs':
-                return (None, None)
+                lines = stdout.decode('utf-8').splitlines()
 
             job = dict()
-            for item in values:
-                logger.debug(f"{args[0]} output {item}")
+            for line in lines:
+                for item in line.split():
+                    logger.debug(f"{args[0]} output {item}")
 
-                # Output is in the form of many key=value pairs, multiple pairs on each line
-                # and multiple lines in the output. Each pair is pulled out of each line and
-                # added to a dictionary
-                for v in values:
-                    bits = v.split('=')
-                    job[bits[0]] = bits[1]
+                    # Output is in the form of many key=value pairs, multiple pairs on each line
+                    # and multiple lines in the output. Each pair is pulled out of each line and
+                    # added to a dictionary.
+                    # Note: In some cases, the value itself may contain white-space. So, if we find
+                    # a key without a value, we consider that key part of the previous value.
+                    bits = item.split('=', 1)
+                    if len(bits) == 1:
+                        job[key] += ' ' + bits[0]
+                    else:
+                        key = bits[0]
+                        job[key] = bits[1]
 
             state = job['JobState']
             try:
@@ -158,14 +165,25 @@ class SlurmBatchSystem(AbstractGridEngineBatchSystem):
         Implementation-specific helper methods
         """
 
-        def prepareSbatch(self, cpu, mem, jobID, jobName):
+        def prepareSbatch(self,
+                          cpu: int,
+                          mem: int,
+                          jobID: int,
+                          jobName: str,
+                          job_environment: Optional[Dict[str, str]]) -> List[str]:
+
             #  Returns the sbatch command line before the script to run
             sbatch_line = ['sbatch', '-J', 'toil_job_{}_{}'.format(jobID, jobName)]
 
-            if self.boss.environment:
+            environment = {}
+            environment.update(self.boss.environment)
+            if job_environment:
+                environment.update(job_environment)
+
+            if environment:
                 argList = []
 
-                for k, v in self.boss.environment.items():
+                for k, v in environment.items():
                     quoted_value = quote(os.environ[k] if v is None else v)
                     argList.append('{}={}'.format(k, quoted_value))
 
@@ -177,8 +195,8 @@ class SlurmBatchSystem(AbstractGridEngineBatchSystem):
             if cpu is not None:
                 sbatch_line.append(f'--cpus-per-task={math.ceil(cpu)}')
 
-            stdoutfile = self.boss.formatStdOutErrPath(jobID, 'slurm', '%j', 'std_output')
-            stderrfile = self.boss.formatStdOutErrPath(jobID, 'slurm', '%j', 'std_error')
+            stdoutfile: str = self.boss.formatStdOutErrPath(jobID, '%j', 'out')
+            stderrfile: str = self.boss.formatStdOutErrPath(jobID, '%j', 'err')
             sbatch_line.extend(['-o', stdoutfile, '-e', stderrfile])
 
             # "Native extensions" for SLURM (see DRMAA or SAGA)
@@ -229,26 +247,3 @@ class SlurmBatchSystem(AbstractGridEngineBatchSystem):
                 # Add a 20% ceiling on the wait duration relative to the scheduler update duration
                 time_value_list.append(math.ceil(time_value*1.2))
         return max(time_value_list)
-
-    @classmethod
-    def obtainSystemConstants(cls):
-        # sinfo -Ne --format '%m,%c'
-        # sinfo arguments:
-        # -N for node-oriented
-        # -h for no header
-        # -e for exact values (e.g. don't return 32+)
-        # --format to get memory, cpu
-        max_cpu = 0
-        max_mem = MemoryString('0')
-        lines = call_command(['sinfo', '-Nhe', '--format', '%m %c']).split('\n')
-        for line in lines:
-            logger.debug("sinfo output %s", line)
-            values = line.split()
-            if len(values) < 2:
-                continue
-            mem, cpu = values
-            max_cpu = max(max_cpu, int(cpu))
-            max_mem = max(max_mem, MemoryString(mem + 'M'))
-        if max_cpu == 0 or max_mem.byteVal() == 0:
-            raise RuntimeError('sinfo did not return memory or cpu info')
-        return max_cpu, max_mem
