@@ -24,10 +24,11 @@ import threading
 import traceback
 from contextlib import contextmanager
 
-import psutil
+import psutil # type: ignore
 
 from toil.lib.exceptions import raise_
-from toil.lib.misc import robust_rmtree
+from toil.lib.io import robust_rmtree
+from typing import Dict, Union, Iterator, Optional, Any
 
 logger = logging.getLogger(__name__)
 
@@ -63,17 +64,17 @@ class ExceptionalThread(threading.Thread):
     """
     exc_info = None
 
-    def run(self):
+    def run(self) -> None:
         try:
             self.tryRun()
         except:
             self.exc_info = sys.exc_info()
             raise
 
-    def tryRun(self):
+    def tryRun(self) -> None:
         super(ExceptionalThread, self).run()
 
-    def join(self, *args, **kwargs):
+    def join(self, *args: Optional[float], **kwargs: Optional[float]) -> None:
         super(ExceptionalThread, self).join(*args, **kwargs)
         if not self.is_alive() and self.exc_info is not None:
             exc_type, exc_value, traceback = self.exc_info
@@ -81,7 +82,7 @@ class ExceptionalThread(threading.Thread):
             raise_(exc_type, exc_value, traceback)
 
 
-def cpu_count():
+def cpu_count() -> Any:
     """
     Get the rounded-up integer number of whole CPUs available.
 
@@ -101,7 +102,7 @@ def cpu_count():
 
     cached = getattr(cpu_count, 'result', None)
     if cached is not None:
-        # We already got a CPU count.    
+        # We already got a CPU count.
         return cached
 
     # Get the fallback answer of all the CPUs on the machine
@@ -134,7 +135,7 @@ def cpu_count():
     except:
         # We can't actually read these cgroup fields. Maybe we are a mac or something.
         logger.debug('Could not inspect cgroup: %s', traceback.format_exc())
-        cgroup_size = float('inf')
+        cgroup_size = float('inf') # type: ignore
 
     # Return the smaller of the actual thread count and the cgroup's limit, minimum 1.
     result = max(1, min(cgroup_size, total_machine_size))
@@ -142,7 +143,7 @@ def cpu_count():
     # Make sure to remember it for the next call
     setattr(cpu_count, 'result', result)
     return result
-    
+
 
 # PIDs are a bad identifier, because they are not shared between containers
 # and also may be reused.
@@ -162,9 +163,9 @@ def cpu_count():
 current_process_name_lock = threading.Lock()
 # And a global dict from work directory to name in that work directory.
 # We also have a file descriptor per work directory but it is just leaked.
-current_process_name_for = {}
+current_process_name_for: Dict[str, str] = {}
 
-def collect_process_name_garbage():
+def collect_process_name_garbage() -> None:
     """
     Delete all the process names that point to files that don't exist anymore
     (because the work directory was temporary and got cleaned up). This is
@@ -172,7 +173,7 @@ def collect_process_name_garbage():
 
     Caller must hold current_process_name_lock.
     """
-    
+
     global current_process_name_for
 
     # Collect the workDirs of the missing names to delete them after iterating.
@@ -186,7 +187,7 @@ def collect_process_name_garbage():
     for workDir in missing:
         del current_process_name_for[workDir]
 
-def destroy_all_process_names():
+def destroy_all_process_names() -> None:
     """
     Delete all our process name files because our process is going away.
 
@@ -203,7 +204,7 @@ def destroy_all_process_names():
 # Run the cleanup at exit
 atexit.register(destroy_all_process_names)
 
-def get_process_name(workDir):
+def get_process_name(workDir: str) -> str:
     """
     Return the name of the current process. Like a PID but visible between
     containers on what to Toil appears to be a node.
@@ -240,13 +241,13 @@ def get_process_name(workDir):
         current_process_name_for[workDir] = os.path.basename(nameFileName)
 
         # Return the basename
-        return current_process_name_for[workDir] 
+        return current_process_name_for[workDir]
 
         # TODO: we leave the file open forever. We might need that in order for
         # it to stay locked while we are alive.
 
 
-def process_name_exists(workDir, name):
+def process_name_exists(workDir: str, name: str) -> bool:
     """
     Return true if the process named by the given name (from process_name) exists, and false otherwise.
 
@@ -260,7 +261,7 @@ def process_name_exists(workDir, name):
 
     global current_process_name_lock
     global current_process_name_for
-    
+
     with current_process_name_lock:
         if current_process_name_for.get(workDir, None) == name:
             # We are asking about ourselves. We are alive.
@@ -272,7 +273,7 @@ def process_name_exists(workDir, name):
         # If the file is gone, the process can't exist.
         return False
 
-    
+
     nameFD = None
     try:
         # Otherwise see if we can lock it shared, for which we need an FD, but
@@ -300,46 +301,46 @@ def process_name_exists(workDir, name):
                 os.close(nameFD)
             except:
                 pass
-    
+
 # Similar to the process naming system above, we define a global mutex system
 # for critical sections, based just around file locks.
 @contextmanager
-def global_mutex(workDir, mutex):
+def global_mutex(workDir: str, mutex: str) -> Iterator[None]:
     """
     Context manager that locks a mutex. The mutex is identified by the given
     name, and scoped to the given directory. Works across all containers that
     have access to the given diectory. Mutexes held by dead processes are
     automatically released.
-    
+
     Only works between processes, NOT between threads.
-    
+
     :param str workDir: The Toil work directory. Defines the shared namespace.
     :param str mutex: Mutex to lock. Must be a permissible path component.
     """
-    
+
     # Define a filename
     lock_filename = os.path.join(workDir, 'toil-mutex-' + mutex)
-    
+
     logger.debug('PID %d acquiring mutex %s', os.getpid(), lock_filename)
-    
+
     # We can't just create/open and lock a file, because when we clean up
     # there's a race where someone can open the file before we unlink it and
     # get a lock on the deleted file.
-    
+
     while True:
         # Try to create the file, ignoring if it exists or not.
         fd = os.open(lock_filename, os.O_CREAT | os.O_WRONLY)
-        
+
         # Wait until we can exclusively lock it.
         fcntl.lockf(fd, fcntl.LOCK_EX)
-        
+
         # Holding the lock, make sure we are looking at the same file on disk still.
         fd_stats = os.fstat(fd)
         try:
             path_stats = os.stat(lock_filename)
         except FileNotFoundError:
-            path_stats = None
-        
+            path_stats = None # type: ignore
+
         if path_stats is None or fd_stats.st_dev != path_stats.st_dev or fd_stats.st_ino != path_stats.st_ino:
             # The file we have a lock on is not the file linked to the name (if
             # any). This usually happens, because before someone releases a
@@ -353,7 +354,7 @@ def global_mutex(workDir, mutex):
             # hold the lock, nobody will be deleting it or can be in the
             # process of deleting it. Stop contending; we have the mutex.
             break
-   
+
     try:
         # When we have it, do the thing we are protecting.
         logger.debug('PID %d now holds mutex %s', os.getpid(), lock_filename)
@@ -369,110 +370,110 @@ def global_mutex(workDir, mutex):
         # get the worthless lock on the now-unlinked file. We have to do some
         # stat gymnastics above to work around this.
         os.close(fd)
-        
+
 
 class LastProcessStandingArena:
     """
     Class that lets a bunch of processes detect and elect a last process
     standing.
-    
+
     Process enter and leave (sometimes due to sudden existence failure). We
     guarantee that the last process to leave, if it leaves properly, will get a
     chance to do some cleanup. If new processes try to enter during the
     cleanup, they will be delayed until after the cleanup has happened and the
     previous "last" process has finished leaving.
-    
+
     The user is responsible for making sure you always leave if you enter!
     Consider using a try/finally; this class is not a context manager.
     """
 
-    def __init__(self, workDir, name):
+    def __init__(self, workDir: str, name: str) -> None:
         """
         Connect to the arena specified by the given workDir and name.
-        
+
         Any process that can access workDir, in any container, can connect to
         the arena. Many arenas can be active with different names.
-        
+
         Doesn't enter or leave the arena.
-        
+
         :param str workDir: The Toil work directory. Defines the shared namespace.
         :param str name: Name of the arena. Must be a permissible path component.
         """
-        
+
         # Save the workDir which namespaces everything
         self.workDir = workDir
-        
+
         # We need a mutex name to allow only one process to be entering or
         # leaving at a time.
         self.mutex = name + '-arena-lock'
-        
+
         # We need a way to track who is actually in, and who was in but died.
         # So everybody gets a locked file (again).
         # TODO: deduplicate with the similar logic for process names, and also
         # deferred functions.
         self.lockfileDir = os.path.join(workDir, name + '-arena-members')
-        
+
         # When we enter the arena, we fill this in with the FD of the locked
         # file that represents our presence.
         self.lockfileFD = None
         # And we fill this in with the file name
         self.lockfileName = None
-        
-    def enter(self):
+
+    def enter(self) -> None:
         """
         This process is entering the arena. If cleanup is in progress, blocks
         until it is finished.
-        
+
         You may not enter the arena again before leaving it.
         """
-       
+
         logger.debug('Joining arena %s', self.lockfileDir)
-       
+
         # Make sure we're not in it already.
         assert self.lockfileName is None
         assert self.lockfileFD is None
-        
+
         with global_mutex(self.workDir, self.mutex):
             # Now nobody else should also be trying to join or leave.
-            
+
             try:
                 # Make sure the lockfile directory exists.
                 os.mkdir(self.lockfileDir)
             except FileExistsError:
                 pass
-            
+
             # Make ourselves a file in it and lock it to prove we are alive.
-            self.lockfileFD, self.lockfileName = tempfile.mkstemp(dir=self.lockfileDir)
+            self.lockfileFD, self.lockfileName = tempfile.mkstemp(dir=self.lockfileDir) # type: ignore
             # Nobody can see it yet, so lock it right away
-            fcntl.lockf(self.lockfileFD, fcntl.LOCK_EX)
-            
+            fcntl.lockf(self.lockfileFD, fcntl.LOCK_EX) # type: ignore
+
             # Now we're properly in, so release the global mutex
-            
+
         logger.debug('Now in arena %s', self.lockfileDir)
-        
-    def leave(self):
+
+    def leave(self) -> Iterator[bool]:
         """
         This process is leaving the arena. If this process happens to be the
         last process standing, yields something, with other processes blocked
         from joining the arena until the loop body completes and the process
         has finished leaving. Otherwise, does not yield anything.
-        
+
         Should be used in a loop:
-        
+
             for _ in arena.leave():
                 # If we get here, we were the last process. Do the cleanup
                 pass
         """
-        
+
         # Make sure we're in it to start.
         assert self.lockfileName is not None
         assert self.lockfileFD is not None
-        
+
         logger.debug('Leaving arena %s', self.lockfileDir)
-        
+
         with global_mutex(self.workDir, self.mutex):
             # Now nobody else should also be trying to join or leave.
-            
+
             # Take ourselves out.
             try:
                 os.unlink(self.lockfileName)
@@ -482,11 +483,11 @@ class LastProcessStandingArena:
             fcntl.lockf(self.lockfileFD, fcntl.LOCK_UN)
             os.close(self.lockfileFD)
             self.lockfileFD = None
-            
+
             for item in os.listdir(self.lockfileDir):
                 # There is someone claiming to be here. Are they alive?
                 full_path = os.path.join(self.lockfileDir, item)
-                
+
                 fd = os.open(full_path, os.O_RDONLY)
                 try:
                     fcntl.lockf(fd, fcntl.LOCK_SH | fcntl.LOCK_NB)
@@ -506,15 +507,15 @@ class LastProcessStandingArena:
                 # the global mutex, so we are the Last Process Standing.
                 logger.debug('We are the Last Process Standing in arena %s', self.lockfileDir)
                 yield True
-                
+
                 try:
                     # Delete the arena directory so as to leave nothing behind.
                     os.rmdir(self.lockfileDir)
                 except:
                     logger.warning('Could not clean up arena %s completely: %s',
                                    self.lockfileDir, traceback.format_exc())
-            
+
             # Now we're done, whether we were the last one or not, and can
             # release the mutex.
-            
+
         logger.debug('Now out of arena %s', self.lockfileDir)
