@@ -14,12 +14,16 @@
 import datetime
 import logging
 import os
+import socket
 from collections import namedtuple
 from operator import attrgetter
 from statistics import mean, stdev
+from typing import List, Optional
 from urllib.error import URLError
 from urllib.request import urlopen
 
+from toil.lib.aws import zone_to_region
+
 logger = logging.getLogger(__name__)
 
 ZoneTuple = namedtuple('ZoneTuple', ['name', 'price_deviation'])
@@ -38,49 +42,67 @@ def running_on_ec2():
     try:
         urlopen('http://169.254.169.254/latest/dynamic/instance-identity/document', timeout=1)
         return True
-    except URLError:
+    except (URLError, socket.timeout):
         return False
 
-
-def zone_to_region(zone):
-    """Get a region (e.g. us-west-2) from a zone (e.g. us-west-1c)."""
-    from toil.lib.context import Context
-    return Context.availability_zone_re.match(zone).group(1)
-
-
 def get_current_aws_region():
-    aws_zone = get_current_aws_zone()
+    aws_zone = get_best_aws_zone()
     return zone_to_region(aws_zone) if aws_zone else None
 
 
-def get_current_aws_zone(spotBid=None, nodeType=None, ctx=None):
-    try:
-        import boto
-        from boto.utils import get_instance_metadata
-    except ImportError:
-        return None
-
+def get_best_aws_zone(spotBid=None, nodeType=None, boto2_ec2=None, zone_options: List[str] = None) -> Optional[str]:
+    """
+    Get the right AWS zone to use.
+    
+    Reports the TOIL_AWS_ZONE environment variable if set.
+    
+    Otherwise, if we are running on EC2, reports the zone we are running in.
+    
+    Otherwise, if a spot bid, node type, and Boto2 EC2 connection are
+    specified, picks a zone where instances are easy to buy from the zones in
+    the region of the Boto2 connection. These parameters must always be
+    specified together, or not at all.
+    
+    In this case, zone_options can be used to restrict to a subset of the zones
+    in the region.
+    
+    Finally, if a default region is configured in Boto 2, chooses a zone in
+    that region.
+    
+    Returns None if no method can produce a zone to use.
+    """
     zone = os.environ.get('TOIL_AWS_ZONE', None)
     if not zone and running_on_ec2():
         try:
+            import boto
+            from boto.utils import get_instance_metadata
             zone = get_instance_metadata()['placement']['availability-zone']
-        except KeyError:
+        except (KeyError, ImportError):
             pass
     if not zone and spotBid:
         # if spot bid is present, all the other parameters must be as well
-        assert bool(spotBid) == bool(nodeType) == bool(ctx)
+        assert bool(spotBid) == bool(nodeType) == bool(boto2_ec2)
         # if the zone is unset and we are using the spot market, optimize our
         # choice based on the spot history
-        return optimize_spot_bid(ctx=ctx, instance_type=nodeType, spot_bid=float(spotBid))
+        
+        if zone_options is None:
+            # We can use all the zones in the region
+            zone_options = [z.name for z in boto2_ec2.get_all_zones()]
+        
+        return optimize_spot_bid(boto2_ec2, instance_type=nodeType, spot_bid=float(spotBid), zone_options=zone_options)
     if not zone:
-        zone = boto.config.get('Boto', 'ec2_region_name')
-        if zone is not None:
-            zone += 'a'  # derive an availability zone in the region
+        try:
+            import boto
+            zone = boto.config.get('Boto', 'ec2_region_name')
+            if zone is not None:
+                zone += 'a'  # derive an availability zone in the region
+        except ImportError:
+            pass
 
     return zone
 
 
-def choose_spot_zone(zones, bid, spot_history):
+def choose_spot_zone(zones: List[str], bid: float, spot_history: List['boto.ec2.spotpricehistory.SpotPriceHistory']) -> str:
     """
     Returns the zone to put the spot request based on, in order of priority:
 
@@ -88,17 +110,11 @@ def choose_spot_zone(zones, bid, spot_history):
 
        2) zones with the most stable price
 
-    :param list[boto.ec2.zone.Zone] zones:
-    :param float bid:
-    :param list[boto.ec2.spotpricehistory.SpotPriceHistory] spot_history:
-
-    :rtype: str
     :return: the name of the selected zone
 
     >>> from collections import namedtuple
     >>> FauxHistory = namedtuple('FauxHistory', ['price', 'availability_zone'])
-    >>> ZoneTuple = namedtuple('ZoneTuple', ['name'])
-    >>> zones = [ZoneTuple('us-west-2a'), ZoneTuple('us-west-2b')]
+    >>> zones = ['us-west-2a', 'us-west-2b']
     >>> spot_history = [FauxHistory(0.1, 'us-west-2a'), \
                         FauxHistory(0.2, 'us-west-2a'), \
                         FauxHistory(0.3, 'us-west-2b'), \
@@ -125,27 +141,28 @@ def choose_spot_zone(zones, bid, spot_history):
     # standard deviation values.
     markets_under_bid, markets_over_bid = [], []
     for zone in zones:
-        zone_histories = [zone_history for zone_history in spot_history if zone_history.availability_zone == zone.name]
+        zone_histories = [zone_history for zone_history in spot_history if zone_history.availability_zone == zone]
         if zone_histories:
             price_deviation = stdev([history.price for history in zone_histories])
             recent_price = zone_histories[0].price
         else:
             price_deviation, recent_price = 0.0, bid
-        zone_tuple = ZoneTuple(name=zone.name, price_deviation=price_deviation)
+        zone_tuple = ZoneTuple(name=zone, price_deviation=price_deviation)
         (markets_over_bid, markets_under_bid)[recent_price < bid].append(zone_tuple)
 
     return min(markets_under_bid or markets_over_bid, key=attrgetter('price_deviation')).name
 
 
-def optimize_spot_bid(ctx, instance_type, spot_bid):
+def optimize_spot_bid(boto2_ec2, instance_type, spot_bid, zone_options: List[str]):
     """
     Check whether the bid is sane and makes an effort to place the instance in a sensible zone.
+    
+    :param zone_options: The collection of allowed zones to consider, within the region associated with the Boto2 connection.
     """
-    spot_history = _get_spot_history(ctx, instance_type)
+    spot_history = _get_spot_history(boto2_ec2, instance_type)
     if spot_history:
         _check_spot_bid(spot_bid, spot_history)
-    zones = ctx.ec2.get_all_zones()
-    most_stable_zone = choose_spot_zone(zones, spot_bid, spot_history)
+    most_stable_zone = choose_spot_zone(zone_options, spot_bid, spot_history)
     logger.debug("Placing spot instances in zone %s.", most_stable_zone)
     return most_stable_zone
 
@@ -183,7 +200,7 @@ def _check_spot_bid(spot_bid, spot_history):
                  "spot price ($ %f) over the last week", spot_bid, average)
 
 
-def _get_spot_history(ctx, instance_type):
+def _get_spot_history(boto2_ec2, instance_type):
     """
     Returns list of 1,000 most recent spot market data points represented as SpotPriceHistory
     objects. Note: The most recent object/data point will be first in the list.
@@ -192,8 +209,8 @@ def _get_spot_history(ctx, instance_type):
     """
 
     one_week_ago = datetime.datetime.now() - datetime.timedelta(days=7)
-    spot_data = ctx.ec2.get_spot_price_history(start_time=one_week_ago.isoformat(),
-                                               instance_type=instance_type,
-                                               product_description="Linux/UNIX")
+    spot_data = boto2_ec2.get_spot_price_history(start_time=one_week_ago.isoformat(),
+                                                 instance_type=instance_type,
+                                                 product_description="Linux/UNIX")
     spot_data.sort(key=attrgetter("timestamp"), reverse=True)
     return spot_data
