@@ -15,6 +15,7 @@ import fcntl
 import itertools
 import logging
 import os
+import stat
 import subprocess
 import sys
 import tempfile
@@ -25,18 +26,24 @@ from inspect import getsource
 from textwrap import dedent
 from unittest import skipIf
 
-from toil.batchSystems.abstractBatchSystem import (BatchSystemSupport,
+from toil.batchSystems.abstractBatchSystem import (AbstractBatchSystem,
+                                                   BatchSystemSupport,
                                                    InsufficientSystemResources)
 # Don't import any batch systems here that depend on extras
-# in order to import properly. Import them later, in tests 
+# in order to import properly. Import them later, in tests
 # protected by annotations.
 from toil.batchSystems.mesos.test import MesosTestSupport
 from toil.batchSystems.parasol import ParasolBatchSystem
+from toil.batchSystems.registry import (BATCH_SYSTEM_FACTORY_REGISTRY,
+                                        BATCH_SYSTEMS,
+                                        single_machine_batch_system_factory,
+                                        addBatchSystemFactory)
 from toil.test.batchSystems.parasolTestSupport import ParasolTestSupport
 from toil.batchSystems.singleMachine import SingleMachineBatchSystem
-from toil.common import Config
+from toil.common import Config, Toil
 from toil.job import Job, JobDescription
 from toil.lib.threading import cpu_count
+from toil.lib.retry import retry_flaky_test
 from toil.test import (ToilTest,
                        needs_aws_s3,
                        needs_fetchable_appliance,
@@ -78,10 +85,7 @@ class hidden(object):
         """
 
         @abstractmethod
-        def createBatchSystem(self):
-            """
-            :rtype: AbstractBatchSystem
-            """
+        def createBatchSystem(self) -> AbstractBatchSystem:
             raise NotImplementedError
 
         def supportsWallTime(self):
@@ -110,14 +114,14 @@ class hidden(object):
             :rtype: toil.common.Config
             """
             return self.createConfig()
-            
+
         def _mockJobDescription(self, jobStoreID=None, command=None, **kwargs):
             """
             Create a mock-up JobDescription with the given ID, command, and other parameters.
             """
-           
+
             # TODO: Use a real unittest.Mock? For now we make a real instance and just hack it up.
-            
+
             desc = JobDescription(**kwargs)
             # Normally we can't pass in a command or ID, and the job
             # serialization logic takes care of filling them in. We set them
@@ -126,7 +130,7 @@ class hidden(object):
                 desc.command = command
             if jobStoreID is not None:
                 desc.jobStoreID = jobStoreID
-                
+
             return desc
 
         @classmethod
@@ -143,11 +147,12 @@ class hidden(object):
         def tearDown(self):
             self.batchSystem.shutdown()
             super(hidden.AbstractBatchSystemTest, self).tearDown()
-        
-        def testAvailableCores(self):
+
+        def test_available_cores(self):
             self.assertTrue(cpu_count() >= numCores)
-        
-        def testRunJobs(self):
+
+        @retry_flaky_test()
+        def test_run_jobs(self):
             jobDesc1 = self._mockJobDescription(command='sleep 1000', jobName='test1', unitName=None,
                                                 jobStoreID='1', requirements=defaultRequirements)
             jobDesc2 = self._mockJobDescription(command='sleep 1000', jobName='test2', unitName=None,
@@ -205,17 +210,18 @@ class hidden(object):
 
             # Make sure killBatchJobs can handle jobs that don't exist
             self.batchSystem.killBatchJobs([10])
-        
-        def testSetEnv(self):
-            # Parasol disobeys shell rules and stupidly splits the command at
-            # the space character into arguments before exec'ing it, whether
-            # the space is quoted, escaped or not.
 
+        def test_set_env(self):
+            # Parasol disobeys shell rules and splits the command at the space
+            # character into arguments before exec'ing it, whether the space is
+            # quoted, escaped or not.
+
+            # Start with a relatively safe script
             script_shell = 'if [ "x${FOO}" == "xbar" ] ; then exit 23 ; else exit 42 ; fi'
 
             # Escape the semicolons
-            script_protected = script_shell.replace(';', r'\;') 
-             
+            script_protected = script_shell.replace(';', r'\;')
+
             # Turn into a string which convinces bash to take all args and paste them back together and run them
             command = "bash -c \"\\${@}\" bash eval " + script_protected
             jobDesc4 = self._mockJobDescription(command=command, jobName='test4', unitName=None,
@@ -233,7 +239,29 @@ class hidden(object):
             jobUpdateInfo = self.batchSystem.getUpdatedBatchJob(maxWait=1000)
             self.assertEqual(jobUpdateInfo.exitStatus, 23)
             self.assertEqual(jobUpdateInfo.jobID, job5)
-        
+
+        def test_set_job_env(self):
+            """ Test the mechanism for setting per-job environment variables to batch system jobs."""
+            script = 'if [ "x${FOO}" == "xbar" ] ; then exit 23 ; else exit 42 ; fi'
+            command = "bash -c \"\\${@}\" bash eval " + script.replace(';', r'\;')
+
+            # Issue a job with a job environment variable
+            job_desc_6 = self._mockJobDescription(command=command, jobName='test6', unitName=None,
+                                                  jobStoreID='6', requirements=defaultRequirements)
+            job6 = self.batchSystem.issueBatchJob(job_desc_6, job_environment={
+                'FOO': 'bar'
+            })
+            job_update_info = self.batchSystem.getUpdatedBatchJob(maxWait=1000)
+            self.assertEqual(job_update_info.exitStatus, 23)  # this should succeed
+            self.assertEqual(job_update_info.jobID, job6)
+            # Now check that the environment variable doesn't exist for other jobs
+            job_desc_7 = self._mockJobDescription(command=command, jobName='test7', unitName=None,
+                                                  jobStoreID='7', requirements=defaultRequirements)
+            job7 = self.batchSystem.issueBatchJob(job_desc_7)
+            job_update_info = self.batchSystem.getUpdatedBatchJob(maxWait=1000)
+            self.assertEqual(job_update_info.exitStatus, 42)
+            self.assertEqual(job_update_info.jobID, job7)
+
         def testCheckResourceRequest(self):
             if isinstance(self.batchSystem, BatchSystemSupport):
                 checkResourceRequest = self.batchSystem.checkResourceRequest
@@ -250,7 +278,7 @@ class hidden(object):
                 self.assertRaises(AssertionError, checkResourceRequest,
                                   memory=10, cores=None, disk=1000)
                 checkResourceRequest(memory=10, cores=1, disk=100)
-       
+
         def testScalableBatchSystem(self):
             # If instance of scalable batch system
             pass
@@ -283,6 +311,14 @@ class hidden(object):
                 time.sleep(1)
             return runningIDs
 
+        def testAddBatchSystemFactory(self):
+            def test_batch_system_factory():
+                return SingleMachineBatchSystem
+
+            addBatchSystemFactory('testBatchSystem', test_batch_system_factory)
+            assert ('testBatchSystem', test_batch_system_factory) in BATCH_SYSTEM_FACTORY_REGISTRY.items()
+            assert 'testBatchSystem' in BATCH_SYSTEMS
+
     class AbstractBatchSystemJobTest(ToilTest, metaclass=ABCMeta):
         """
         An abstract base class for batch system tests that use a full Toil workflow rather
@@ -338,10 +374,33 @@ class hidden(object):
                 for _ in range(self.cpuCount):
                     root.addFollowOn(Job.wrapFn(measureConcurrency, counterPath, self.sleepTime,
                                                 cores=coresPerJob, memory='1M', disk='1Mi'))
-                Job.Runner.startToil(root, options)
+                with Toil(options) as toil:
+                    toil.start(root)
                 _, maxValue = getCounters(counterPath)
                 self.assertEqual(maxValue, self.cpuCount // coresPerJob)
 
+        def test_omp_threads(self):
+            """
+            Test if the OMP_NUM_THREADS env var is set correctly based on jobs.cores.
+            """
+            test_cases = {
+                # mapping of the number of cores to the OMP_NUM_THREADS value
+                0.1: "1",
+                1: "1",
+                2: "2"
+            }
+
+            temp_dir = self._createTempDir()
+            options = self.getOptions(temp_dir)
+
+            for cores, expected_omp_threads in test_cases.items():
+                if os.environ.get('OMP_NUM_THREADS'):
+                    expected_omp_threads = os.environ.get('OMP_NUM_THREADS')
+                    logger.info(f"OMP_NUM_THREADS is set.  Using OMP_NUM_THREADS={expected_omp_threads} instead.")
+                with Toil(options) as toil:
+                    output = toil.start(Job.wrapFn(get_omp_threads, memory='1Mi', cores=cores, disk='1Mi'))
+                self.assertEqual(output, expected_omp_threads)
+
     class AbstractGridEngineBatchSystemTest(AbstractBatchSystemTest):
         """
         An abstract class to reduce redundancy between Grid Engine, Slurm, and other similar batch
@@ -371,6 +430,7 @@ class KubernetesBatchSystemTest(hidden.AbstractBatchSystemTest):
         return KubernetesBatchSystem(config=self.config,
                                      maxCores=numCores, maxMemory=1e9, maxDisk=2001)
 
+
 @slow
 @needs_mesos
 class MesosBatchSystemTest(hidden.AbstractBatchSystemTest, MesosTestSupport):
@@ -378,6 +438,7 @@ class MesosBatchSystemTest(hidden.AbstractBatchSystemTest, MesosTestSupport):
     Tests against the Mesos batch system
     """
 
+    @classmethod
     def createConfig(cls):
         """
         needs to set mesosMasterAddress to localhost for testing since the default is now the
@@ -415,19 +476,147 @@ class MesosBatchSystemTest(hidden.AbstractBatchSystemTest, MesosTestSupport):
         # Make sure job is NOT running
         self.assertEqual(set(runningJobIDs), set({}))
 
+
+def write_temp_file(s: str, temp_dir: str) -> str:
+    """
+    Dump a string into a temp file and return its path.
+    """
+    fd, path = tempfile.mkstemp(dir=temp_dir)
+    try:
+        encoded = s.encode('utf-8')
+        assert os.write(fd, encoded) == len(encoded)
+    except:
+        os.unlink(path)
+        raise
+    else:
+        return path
+    finally:
+        os.close(fd)
+
+
 @travis_test
 class SingleMachineBatchSystemTest(hidden.AbstractBatchSystemTest):
     """
     Tests against the single-machine batch system
     """
 
-    def supportsWallTime(self):
+    def supportsWallTime(self) -> bool:
         return True
 
-    def createBatchSystem(self):
+    def createBatchSystem(self) -> AbstractBatchSystem:
         return SingleMachineBatchSystem(config=self.config,
                                         maxCores=numCores, maxMemory=1e9, maxDisk=2001)
 
+    def testProcessEscape(self, hide: bool = False) -> None:
+        """
+        Test to make sure that child processes and their descendants go away
+        when the Toil workflow stops.
+
+        If hide is true, will try and hide the child processes to make them
+        hard to stop.
+        """
+
+        def script() -> None:
+            #!/usr/bin/env python3
+            import fcntl
+            import os
+            import sys
+            import signal
+            import time
+            from typing import Any
+
+            def handle_signal(sig: Any, frame: Any) -> None:
+                sys.stderr.write(f'{os.getpid()} ignoring signal {sig}\n')
+
+            if hasattr(signal, 'valid_signals'):
+                # We can just ask about the signals
+                all_signals = signal.valid_signals()
+            else:
+                # Fish them out by name
+                all_signals = [getattr(signal, n) for n in dir(signal) if n.startswith('SIG') and not n.startswith('SIG_')]
+
+            for sig in all_signals:
+                # Set up to ignore all signals we can and generally be obstinate
+                if sig != signal.SIGKILL and sig != signal.SIGSTOP:
+                    signal.signal(sig, handle_signal)
+
+            if len(sys.argv) > 2:
+                # Instructed to hide
+                if os.fork():
+                    # Try and hide the first process immediately so getting its
+                    # pgid won't work.
+                    sys.exit(0)
+
+            for depth in range(3):
+                # Bush out into a tree of processes
+                os.fork()
+
+            if len(sys.argv) > 1:
+                fd = os.open(sys.argv[1], os.O_RDONLY)
+                fcntl.lockf(fd, fcntl.LOCK_SH)
+
+            sys.stderr.write(f'{os.getpid()} waiting...\n')
+
+            while True:
+                # Wait around forever
+                time.sleep(60)
+
+        # Get a directory where we can safely dump files.
+        temp_dir = self._createTempDir()
+
+        script_path = write_temp_file(self._getScriptSource(script), temp_dir)
+
+        # We will have all the job processes try and lock this file shared while they are alive.
+        lockable_path = write_temp_file('', temp_dir)
+
+        try:
+            command = f'{sys.executable} {script_path} {lockable_path}'
+            if hide:
+                # Tell the children to stop the first child and hide out in the
+                # process group it made.
+                command += ' hide'
+
+            # Start the job
+            self.batchSystem.issueBatchJob(self._mockJobDescription(command=command, jobName='fork',
+                                                                    jobStoreID='1', requirements=defaultRequirements))
+            # Wait
+            time.sleep(10)
+
+            lockfile = open(lockable_path, 'w')
+
+            if not hide:
+                # In hiding mode the job will finish, and the batch system will
+                # clean up after it promptly. In non-hiding mode the job will
+                # stick around until shutdown, so make sure we can see it.
+
+                # Try to lock the file and make sure it fails
+
+                try:
+                    fcntl.lockf(lockfile, fcntl.LOCK_EX | fcntl.LOCK_NB)
+                    assert False, "Should not be able to lock file while job is running"
+                except OSError:
+                    pass
+
+            # Shut down the batch system
+            self.batchSystem.shutdown()
+
+            # After the batch system shuts down, we should be able to get the
+            # lock immediately, because all the children should be gone.
+            fcntl.lockf(lockfile, fcntl.LOCK_EX | fcntl.LOCK_NB)
+            # Then we can release it
+            fcntl.lockf(lockfile, fcntl.LOCK_UN)
+        finally:
+            os.unlink(script_path)
+            os.unlink(lockable_path)
+
+    def testHidingProcessEscape(self):
+        """
+        Test to make sure that child processes and their descendants go away
+        when the Toil workflow stops, even if the job process stops and leaves children.
+        """
+
+        self.testProcessEscape(hide=True)
+
 
 @slow
 class MaxCoresSingleMachineBatchSystemTest(ToilTest):
@@ -437,36 +626,25 @@ class MaxCoresSingleMachineBatchSystemTest(ToilTest):
     """
 
     @classmethod
-    def setUpClass(cls):
+    def setUpClass(cls) -> None:
         super(MaxCoresSingleMachineBatchSystemTest, cls).setUpClass()
         logging.basicConfig(level=logging.DEBUG)
 
-    def setUp(self):
+    def setUp(self) -> None:
         super(MaxCoresSingleMachineBatchSystemTest, self).setUp()
 
-        def writeTempFile(s):
-            fd, path = tempfile.mkstemp()
-            try:
-                encoded = s.encode('utf-8')
-                assert os.write(fd, encoded) == len(encoded)
-            except:
-                os.unlink(path)
-                raise
-            else:
-                return path
-            finally:
-                os.close(fd)
+        temp_dir = self._createTempDir()
 
         # Write initial value of counter file containing a tuple of two integers (i, n) where i
         # is the number of currently executing tasks and n the maximum observed value of i
-        self.counterPath = writeTempFile('0,0')
+        self.counterPath = write_temp_file('0,0', temp_dir)
 
-        def script():
+        def script() -> None:
             import fcntl
             import os
             import sys
             import time
-            def count(delta):
+            def count(delta: int) -> None:
                 """
                 Adjust the first integer value in a file by the given amount. If the result
                 exceeds the second integer value, set the second one to the first.
@@ -498,15 +676,16 @@ class MaxCoresSingleMachineBatchSystemTest(ToilTest):
             else:
                 count(int(sys.argv[2]))
 
-        self.scriptPath = writeTempFile(dedent('\n'.join(getsource(script).split('\n')[1:])))
+        self.scriptPath = write_temp_file(self._getScriptSource(script), temp_dir)
 
-    def tearDown(self):
+    def tearDown(self) -> None:
         os.unlink(self.scriptPath)
         os.unlink(self.counterPath)
 
-    def scriptCommand(self):
+    def scriptCommand(self) -> str:
         return ' '.join([sys.executable, self.scriptPath, self.counterPath])
 
+    @retry_flaky_test()
     def test(self):
         # We'll use fractions to avoid rounding errors. Remember that not every fraction can be
         # represented as a floating point number.
@@ -622,7 +801,7 @@ class ParasolBatchSystemTest(hidden.AbstractBatchSystemTest, ParasolTestSupport)
         config.jobStore = self._createTempDir('jobStore')
         return config
 
-    def createBatchSystem(self):
+    def createBatchSystem(self) -> AbstractBatchSystem:
         memory = int(3e9)
         self._startParasol(numCores=numCores, memory=memory)
 
@@ -687,7 +866,7 @@ class GridEngineBatchSystemTest(hidden.AbstractGridEngineBatchSystemTest):
     Tests against the GridEngine batch system
     """
 
-    def createBatchSystem(self):
+    def createBatchSystem(self) -> AbstractBatchSystem:
         from toil.batchSystems.gridengine import GridEngineBatchSystem
         return GridEngineBatchSystem(config=self.config, maxCores=numCores, maxMemory=1000e9,
                                      maxDisk=1e9)
@@ -707,7 +886,7 @@ class SlurmBatchSystemTest(hidden.AbstractGridEngineBatchSystemTest):
     Tests against the Slurm batch system
     """
 
-    def createBatchSystem(self):
+    def createBatchSystem(self) -> AbstractBatchSystem:
         from toil.batchSystems.slurm import SlurmBatchSystem
         return SlurmBatchSystem(config=self.config, maxCores=numCores, maxMemory=1000e9,
                                 maxDisk=1e9)
@@ -726,7 +905,7 @@ class LSFBatchSystemTest(hidden.AbstractGridEngineBatchSystemTest):
     """
     Tests against the LSF batch system
     """
-    def createBatchSystem(self):
+    def createBatchSystem(self) -> AbstractBatchSystem:
         from toil.batchSystems.lsf import LSFBatchSystem
         return LSFBatchSystem(config=self.config, maxCores=numCores,
                               maxMemory=1000e9, maxDisk=1e9)
@@ -745,7 +924,7 @@ class TorqueBatchSystemTest(hidden.AbstractGridEngineBatchSystemTest):
         config.jobStore = self._createTempDir('jobStore')
         return config
 
-    def createBatchSystem(self):
+    def createBatchSystem(self) -> AbstractBatchSystem:
         from toil.batchSystems.torque import TorqueBatchSystem
         return TorqueBatchSystem(config=self.config, maxCores=numCores, maxMemory=1000e9,
                                      maxDisk=1e9)
@@ -757,6 +936,7 @@ class TorqueBatchSystemTest(hidden.AbstractGridEngineBatchSystemTest):
         for f in glob('toil_job_*.[oe]*'):
             os.unlink(f)
 
+
 @slow
 @needs_htcondor
 class HTCondorBatchSystemTest(hidden.AbstractGridEngineBatchSystemTest):
@@ -764,14 +944,15 @@ class HTCondorBatchSystemTest(hidden.AbstractGridEngineBatchSystemTest):
     Tests against the HTCondor batch system
     """
 
-    def createBatchSystem(self):
+    def createBatchSystem(self) -> AbstractBatchSystem:
         from toil.batchSystems.htcondor import HTCondorBatchSystem
         return HTCondorBatchSystem(config=self.config, maxCores=numCores, maxMemory=1000e9,
-                                       maxDisk=1e9)
+                                   maxDisk=1e9)
 
     def tearDown(self):
         super(HTCondorBatchSystemTest, self).tearDown()
 
+
 @travis_test
 class SingleMachineBatchSystemJobTest(hidden.AbstractBatchSystemJobTest):
     """
@@ -782,6 +963,7 @@ class SingleMachineBatchSystemJobTest(hidden.AbstractBatchSystemJobTest):
         return "single_machine"
 
     @slow
+    @retry_flaky_test()
     def testConcurrencyWithDisk(self):
         """
         Tests that the batch system is allocating disk resources properly
@@ -792,6 +974,9 @@ class SingleMachineBatchSystemJobTest(hidden.AbstractBatchSystemJobTest):
         options.workDir = tempDir
         from toil import physicalDisk
         availableDisk = physicalDisk(options.workDir)
+        logger.info('Testing disk concurrency limits with %s disk space', availableDisk)
+        # More disk might become available by the time Toil starts, so we limit it here
+        options.maxDisk = availableDisk
         options.batchSystem = self.batchSystemName
 
         counterPath = os.path.join(tempDir, 'counter')
@@ -799,15 +984,22 @@ class SingleMachineBatchSystemJobTest(hidden.AbstractBatchSystemJobTest):
         value, maxValue = getCounters(counterPath)
         assert (value, maxValue) == (0, 0)
 
+        half_disk = availableDisk // 2
+        more_than_half_disk = half_disk + 500
+        logger.info('Dividing into parts of %s and %s', half_disk, more_than_half_disk)
+
         root = Job()
         # Physically, we're asking for 50% of disk and 50% of disk + 500bytes in the two jobs. The
         # batchsystem should not allow the 2 child jobs to run concurrently.
         root.addChild(Job.wrapFn(measureConcurrency, counterPath, self.sleepTime, cores=1,
-                                    memory='1M', disk=availableDisk // 2))
+                                 memory='1M', disk=half_disk))
         root.addChild(Job.wrapFn(measureConcurrency, counterPath, self.sleepTime, cores=1,
-                                 memory='1M', disk=(availableDisk // 2) + 500))
+                                 memory='1M', disk=more_than_half_disk))
         Job.Runner.startToil(root, options)
         _, maxValue = getCounters(counterPath)
+
+        logger.info('After run: %s disk space', physicalDisk(options.workDir))
+
         self.assertEqual(maxValue, 1)
 
     @skipIf(SingleMachineBatchSystem.numCores < 4, 'Need at least four cores to run this test')
@@ -970,3 +1162,7 @@ def resetCounters(path):
     with open(path, "w") as f:
         f.write("0,0")
         f.close()
+
+
+def get_omp_threads() -> str:
+    return os.environ['OMP_NUM_THREADS']
