@@ -41,9 +41,11 @@ class ToilDocumentationTest(ToilTest):
         process = subprocess.Popen([python, program, "file:my-jobstore", "--clean=always"],
                                    stdout=subprocess.PIPE, stderr=subprocess.PIPE)
         stdout, stderr = process.communicate()
-        assert process.returncode == 0, stderr
         if isinstance(stdout, bytes):
-            return stdout.decode('utf-8') + ' ' + stderr.decode('utf-8')
+            stdout = stdout.decode('utf-8')
+            stderr = stderr.decode('utf-8')
+        if not process.returncode == 0:
+            raise RuntimeError(stderr)
         return stdout + ' ' + stderr
 
     """Check the exit code and the output"""
@@ -160,6 +162,10 @@ class ToilDocumentationTest(ToilTest):
     def testServices(self):
         self.checkExitCode("tutorial_services.py")
 
+    @travis_test
+    def testStaging(self):
+        self.checkExitCode("tutorial_staging.py")
+
     @unittest.skip('Needs cromwell jar file to run.')
     def testWdlexample(self):
         self.checkExitCode("tutorial_wdlexample.py")
