@@ -24,10 +24,11 @@ import time
 import urllib.parse as urlparse
 import uuid
 from abc import ABCMeta, abstractmethod
-from io import StringIO
+from io import BytesIO
 from itertools import chain, islice
 from queue import Queue
 from threading import Thread
+from typing import Any, Tuple
 from urllib.request import Request, urlopen
 
 import pytest
@@ -40,6 +41,7 @@ from toil.jobStores.abstractJobStore import (NoSuchFileException,
                                              NoSuchJobException)
 from toil.jobStores.fileJobStore import FileJobStore
 from toil.lib.memoize import memoize
+from toil.lib.aws.utils import create_s3_bucket
 from toil.statsAndLogging import StatsAndLogging
 from toil.test import (ToilTest,
                        make_tests,
@@ -224,7 +226,7 @@ class AbstractJobStoreTest(object):
             self.jobstore_initialized.create(childJob)
             job.addChild(childJob.jobStoreID)
             self.jobstore_initialized.update(job)
-            
+
             self.assertEqual(self.jobstore_initialized.load(list(job.allSuccessors())[0]).command, childJob.command)
 
         @travis_test
@@ -288,7 +290,7 @@ class AbstractJobStoreTest(object):
             # Reload parent job on jobstore, "refreshing" the job.
             job1 = jobstore1.load(job1.jobStoreID)
             self.assertEqual([sorted(x) for x in job2.stack], [sorted(x) for x in job1.stack])
-            
+
             # Jobs still shouldn't *actually* be equal, even if their contents are the same.
             self.assertNotEqual(job2, job1)
 
@@ -316,7 +318,7 @@ class AbstractJobStoreTest(object):
             child2 = JobDescription(command='job1',
                                     requirements=self.childJobReqs2,
                                     jobName='test3', unitName='onChild2')
-                                    
+
             # Add children to parent.
             jobstore.assignID(child1)
             jobstore.create(child1)
@@ -325,11 +327,11 @@ class AbstractJobStoreTest(object):
             job.addChild(child1.jobStoreID)
             job.addChild(child2.jobStoreID)
             jobstore.update(job)
-            
+
             # Get it ready to run children
             job.command = None
             jobstore.update(job)
-            
+
             # Go get the children
             childJobs = [jobstore.load(childID) for childID in job.nextSuccessors()]
 
@@ -382,6 +384,44 @@ class AbstractJobStoreTest(object):
             self.assertUrl(jobstore1.getSharedPublicUrl('nonEncrypted'))
             self.assertRaises(NoSuchFileException, jobstore1.getSharedPublicUrl, 'missing')
 
+        def testReadWriteSharedFilesTextMode(self):
+            """Checks if text mode is compatible for shared file streams."""
+            jobstore1 = self.jobstore_initialized
+            jobstore2 = self.jobstore_resumed_noconfig
+
+            bar = 'bar'
+
+            with jobstore1.writeSharedFileStream('foo', encoding='utf-8') as f:
+                f.write(bar)
+
+            with jobstore2.readSharedFileStream('foo', encoding='utf-8') as f:
+                self.assertEqual(bar, f.read())
+
+            with jobstore1.readSharedFileStream('foo', encoding='utf-8') as f:
+                self.assertEqual(bar, f.read())
+
+        def testReadWriteFileStreamTextMode(self):
+            """Checks if text mode is compatible for file streams."""
+            jobstore = self.jobstore_initialized
+            job = self.arbitraryJob()
+            jobstore.assignID(job)
+            jobstore.create(job)
+
+            foo = 'foo'
+            bar = 'bar'
+
+            with jobstore.writeFileStream(job.jobStoreID, encoding='utf-8') as (f, fileID):
+                f.write(foo)
+
+            with jobstore.readFileStream(fileID, encoding='utf-8') as f:
+                self.assertEqual(foo, f.read())
+
+            with jobstore.updateFileStream(fileID, encoding='utf-8') as f:
+                f.write(bar)
+
+            with jobstore.readFileStream(fileID, encoding='utf-8') as f:
+                self.assertEqual(bar, f.read())
+
         @travis_test
         def testPerJobFiles(self):
             """Tests the behavior of files on jobs."""
@@ -465,7 +505,7 @@ class AbstractJobStoreTest(object):
             jobOnJobStore1 = JobDescription(command='job1',
                                             requirements=self.parentJobReqs,
                                             jobName='test1', unitName='onJobStore1')
-            
+
             jobstore1.assignID(jobOnJobStore1)
             jobstore1.create(jobOnJobStore1)
 
@@ -967,7 +1007,7 @@ class AbstractJobStoreTest(object):
             """Test whether readFileStream will deadlock on a partial read."""
             job = self.arbitraryJob()
             self.jobstore_initialized.assignID(job)
-            self.jobstore_initialized.create(job) 
+            self.jobstore_initialized.create(job)
             with self.jobstore_initialized.writeFileStream(job.jobStoreID, cleanup=True) as (f, fileID):
                 # Write enough data to make sure the writer thread
                 # will get blocked on the write. Technically anything
@@ -1145,6 +1185,45 @@ class FileJobStoreTest(AbstractJobStoreTest.Test):
             if original_filestore and os.path.exists(original_filestore):
                 shutil.rmtree(original_filestore)
 
+    @travis_test
+    def test_jobstore_does_not_leak_symlinks(self):
+        """Test that if we link imports into the FileJobStore, we can't get hardlinks to symlinks."""
+
+        temp_dir = None
+        download_dir = None
+
+        try:
+            # Grab a temp directory to make files in. Make sure it's on the
+            # same device as everything else.
+            temp_dir = os.path.abspath(self.namePrefix + '-import')
+            os.mkdir(temp_dir)
+            to_import = os.path.join(temp_dir, 'import-me')
+            with open(to_import, 'w') as f:
+                f.write('test')
+
+            # And a temp directory next to the job store to download to
+            download_dir = os.path.abspath(self.namePrefix + '-dl')
+            os.mkdir(download_dir)
+
+            # Import it as a symlink
+            file_id = self.jobstore_initialized.importFile('file://' + to_import, symlink=True)
+
+            # Take it out as a hard link or copy
+            download_to = os.path.join(download_dir, 'downloaded')
+            self.jobstore_initialized.readFile(file_id, download_to)
+
+            # Make sure it isn't a symlink
+            self.assertFalse(os.path.islink(download_to))
+
+        finally:
+            if temp_dir and os.path.exists(temp_dir):
+                # Clean up temp directory
+                shutil.rmtree(temp_dir)
+            if download_dir and os.path.exists(download_dir):
+                # Clean up download directory
+                shutil.rmtree(download_dir)
+
+
 
 @needs_google
 class GoogleJobStoreTest(AbstractJobStoreTest.Test):
@@ -1168,8 +1247,8 @@ class GoogleJobStoreTest(AbstractJobStoreTest.Test):
             return url
         with open('/dev/urandom', 'rb') as readable:
             contents = str(readable.read(size))
-        GoogleJobStore._writeToUrl(StringIO(contents), urlparse.urlparse(url))
-        return url, hashlib.md5(contents).hexdigest()
+        GoogleJobStore._writeToUrl(BytesIO(bytes(contents, 'utf-8')), urlparse.urlparse(url))
+        return url, hashlib.md5(contents.encode()).hexdigest()
 
     def _hashTestFile(self, url):
         from toil.jobStores.googleJobStore import GoogleJobStore
@@ -1216,9 +1295,9 @@ class AWSJobStoreTest(AbstractJobStoreTest.Test):
         from boto.sdb import connect_to_region
         from toil.jobStores.aws.jobStore import BucketLocationConflictException
         from toil.jobStores.aws.utils import retry_s3
-        from toil.jobStores.aws.jobStore import get_boto3_session
-        externalAWSLocation = 'us-west-1'
+        from toil.jobStores.aws.jobStore import establish_boto3_session
 
+        externalAWSLocation = 'us-west-1'
         for testRegion in 'us-east-1', 'us-west-2':
             # We run this test twice, once with the default s3 server us-east-1 as the test region
             # and once with another server (us-west-2).  The external server is always us-west-1.
@@ -1227,15 +1306,23 @@ class AWSJobStoreTest(AbstractJobStoreTest.Test):
             testJobStoreUUID = str(uuid.uuid4())
             # Create the bucket at the external region
             bucketName = 'domain-test-' + testJobStoreUUID + '--files'
-            client = get_boto3_session().client('s3', region_name=externalAWSLocation)
+            client = establish_boto3_session().client('s3', region_name=externalAWSLocation)
+            resource = establish_boto3_session().resource('s3', region_name=externalAWSLocation)
 
             for attempt in retry_s3(delays=(2, 5, 10, 30, 60), timeout=600):
                 with attempt:
+                    # Create the bucket at the home region
                     client.create_bucket(Bucket=bucketName,
                                          CreateBucketConfiguration={'LocationConstraint': externalAWSLocation})
 
-            options = Job.Runner.getDefaultOptions('aws:' + testRegion + ':domain-test-' +
-                                                   testJobStoreUUID)
+            owner_tag = os.environ.get('TOIL_OWNER_TAG')
+            if owner_tag:
+                for attempt in retry_s3(delays=(1, 1, 2, 4, 8, 16), timeout=33):
+                    with attempt:
+                        bucket_tagging = resource.BucketTagging(bucketName)
+                        bucket_tagging.put(Tagging={'TagSet': [{'Key': 'Owner', 'Value': owner_tag}]})
+
+            options = Job.Runner.getDefaultOptions('aws:' + testRegion + ':domain-test-' + testJobStoreUUID)
             options.logLevel = 'DEBUG'
             try:
                 with Toil(options) as toil:
@@ -1301,20 +1388,58 @@ class AWSJobStoreTest(AbstractJobStoreTest.Test):
         self.assertEqual(jobsInJobStore, [str(overlargeJob)])
         jobstore.delete(overlargeJob.jobStoreID)
 
+    def testMultiThreadImportFile(self) -> None:
+        """ Tests that importFile is thread-safe."""
+
+        from concurrent.futures.thread import ThreadPoolExecutor
+        from toil.lib.threading import cpu_count
+
+        threads: Tuple[int, ...] = (2, cpu_count()) if cpu_count() > 2 else (2, )
+        num_of_files: int = 5
+        size: int = 1 << 16 + 1
+
+        # The string in otherCls() is arbitrary as long as it returns a class that has access
+        # to ._externalStore() and ._prepareTestFile()
+        other: AbstractJobStoreTest.Test = AWSJobStoreTest('testSharedFiles')
+        store: Any = other._externalStore()
+
+        # prepare test files to import
+        logger.debug(f'Preparing {num_of_files} test files for testMultiThreadImportFile().')
+        test_files = [other._prepareTestFile(store, size) for _ in range(num_of_files)]
+
+        for thread_count in threads:
+            with self.subTest(f'Testing threaded importFile with "{thread_count}" threads.'):
+                results = []
+
+                with ThreadPoolExecutor(max_workers=thread_count) as executor:
+                    for url, expected_md5 in test_files:
+                        # run jobStore.importFile() asynchronously
+                        future = executor.submit(self.jobstore_initialized.importFile, url)
+                        results.append((future, expected_md5))
+
+                self.assertEqual(len(results), num_of_files)
+
+                for future, expected_md5 in results:
+                    file_id = future.result()
+                    self.assertIsInstance(file_id, FileID)
+
+                    with self.jobstore_initialized.readFileStream(file_id) as f:
+                        self.assertEqual(hashlib.md5(f.read()).hexdigest(), expected_md5)
+
     def _prepareTestFile(self, bucket, size=None):
         from toil.jobStores.aws.utils import retry_s3
 
-        fileName = 'testfile_%s' % uuid.uuid4()
-        url = 's3://%s/%s' % (bucket.name, fileName)
+        file_name = 'testfile_%s' % uuid.uuid4()
+        url = 's3://%s/%s' % (bucket.name, file_name)
         if size is None:
             return url
         with open('/dev/urandom', 'rb') as readable:
             for attempt in retry_s3():
                 with attempt:
-                    bucket.put_object(Key=fileName, Body=str(readable.read(size)))
-        return url, hashlib.md5(bucket.Object(fileName).get().get('Body').read()).hexdigest()
+                    bucket.put_object(Key=file_name, Body=str(readable.read(size)))
+        return url, hashlib.md5(bucket.Object(file_name).get().get('Body').read()).hexdigest()
 
-    def _hashTestFile(self, url):
+    def _hashTestFile(self, url: str) -> str:
         from toil.jobStores.aws.jobStore import AWSJobStore
         key = AWSJobStore._getObjectForUrl(urlparse.urlparse(url), existing=True)
         contents = key.get().get('Body').read()
@@ -1323,14 +1448,17 @@ class AWSJobStoreTest(AbstractJobStoreTest.Test):
     def _createExternalStore(self):
         """A S3.Bucket instance is returned"""
         from toil.jobStores.aws.utils import retry_s3
-        from toil.jobStores.aws.utils import region_to_bucket_location
-        from toil.jobStores.aws.jobStore import get_boto3_session
-        resource = get_boto3_session().resource('s3', region_name=self.awsRegion())
-        bucket = resource.Bucket('import-export-test-%s' % uuid.uuid4())
+        from toil.jobStores.aws.jobStore import establish_boto3_session
+
+        resource = establish_boto3_session().resource(
+            "s3", region_name=self.awsRegion()
+        )
+        bucket_name = f"import-export-test-{uuid.uuid4()}"
+        location = self.awsRegion()
 
         for attempt in retry_s3():
             with attempt:
-                bucket.create(CreateBucketConfiguration={'LocationConstraint': region_to_bucket_location(self.awsRegion())})
+                bucket = create_s3_bucket(resource, bucket_name, location)
                 bucket.wait_until_exists()
                 return bucket
 
