@@ -22,6 +22,7 @@ from collections import defaultdict
 from contextlib import contextmanager
 from queue import Empty, Queue
 from threading import Event, Thread
+from typing import Set, List, Optional, Tuple
 
 from mock import MagicMock
 
@@ -30,7 +31,8 @@ from toil.batchSystems.abstractBatchSystem import (AbstractBatchSystem,
                                                    NodeInfo)
 from toil.common import Config, defaultTargetTime
 from toil.job import JobDescription
-from toil.lib.humanize import human2bytes as h2b
+from toil.lib.conversions import human2bytes as h2b
+from toil.provisioners import parse_node_types
 from toil.provisioners.abstractProvisioner import AbstractProvisioner, Shape
 from toil.provisioners.clusterScaler import (BinPackedFit,
                                              ClusterScaler,
@@ -68,9 +70,9 @@ t2_micro = Shape(wallTime=3600,
 
 class BinPackingTest(ToilTest):
     def setUp(self):
-        self.nodeShapes = [c4_8xlarge_preemptable, r3_8xlarge]
-        self.bpf = BinPackedFit(self.nodeShapes)
-    
+        self.node_shapes_for_testing = [c4_8xlarge_preemptable, r3_8xlarge]
+        self.bpf = BinPackedFit(self.node_shapes_for_testing)
+
     @travis_test
     def testPackingOneShape(self):
         """Pack one shape and check that the resulting reservations look sane."""
@@ -92,7 +94,7 @@ class BinPackingTest(ToilTest):
                                  cores=36,
                                  disk=h2b('100G'),
                                  preemptable=True)]])
-    
+
     @travis_test
     def testSorting(self):
         """
@@ -106,7 +108,7 @@ class BinPackingTest(ToilTest):
                              t2_micro, t2_micro, t2_micro,
                              c4_8xlarge, c4_8xlarge, c4_8xlarge,
                              r3_8xlarge, r3_8xlarge, r3_8xlarge]
-    
+
     @travis_test
     def testAddingInitialNode(self):
         """Pack one shape when no nodes are available and confirm that we fit one node properly."""
@@ -126,7 +128,7 @@ class BinPackingTest(ToilTest):
                                  cores=36,
                                  disk=h2b('100G'),
                                  preemptable=True)]])
-    
+
     @travis_test
     def testLowTargetTime(self):
         """
@@ -148,7 +150,7 @@ class BinPackingTest(ToilTest):
                                               jobTime=300,
                                               globalTargetTime=0)
         self.assertEqual(allocation, {t2_micro: 1000})
-    
+
     @travis_test
     def testHighTargetTime(self):
         """
@@ -169,14 +171,14 @@ class BinPackingTest(ToilTest):
                                               jobTime=300,
                                               globalTargetTime=3600)
         self.assertEqual(allocation, {t2_micro: 84})
-    
+
     @travis_test
     def testZeroResourceJobs(self):
         """
         Test that jobs requiring zero cpu/disk/mem pack first, regardless of targetTime.
 
         Disk/cpu/mem packing is prioritized first, so we set job resource reqs so that each
-        t2.micro (1 cpu/8G disk/1G RAM) can run a seemingly infinite number of jobs with its 
+        t2.micro (1 cpu/8G disk/1G RAM) can run a seemingly infinite number of jobs with its
         resources.
 
         Since all jobs should pack cpu/disk/mem-wise on a t2.micro, we expect only one t2.micro to
@@ -188,7 +190,7 @@ class BinPackingTest(ToilTest):
                                               jobTime=300,
                                               globalTargetTime=0)
         self.assertEqual(allocation, {t2_micro: 1})
-    
+
     @travis_test
     def testLongRunningJobs(self):
         """
@@ -212,8 +214,8 @@ class BinPackingTest(ToilTest):
         """Test packing 1000 jobs on t2.micros.  Depending on the targetTime and resources,
         these should pack differently.
         """
-        nodeShapes = [t2_micro]
-        bpf = BinPackedFit(nodeShapes, targetTime=globalTargetTime)
+        node_shapes_for_testing = [t2_micro]
+        bpf = BinPackedFit(node_shapes_for_testing, targetTime=globalTargetTime)
 
         for _ in range(1000):
             bpf.addJobShape(Shape(wallTime=jobTime,
@@ -222,7 +224,7 @@ class BinPackingTest(ToilTest):
                                    disk=jobDisk,
                                    preemptable=False))
         return bpf.getRequiredNodes()
-    
+
     @travis_test
     def testPathologicalCase(self):
         """Test a pathological case where only one node can be requested to fit months' worth of jobs.
@@ -248,7 +250,7 @@ class BinPackingTest(ToilTest):
                                        preemptable=False))
         # Hopefully we didn't assign just one node to cover all those jobs.
         self.assertNotEqual(self.bpf.getRequiredNodes(), {r3_8xlarge: 1, c4_8xlarge_preemptable: 0})
-    
+
     @travis_test
     def testJobTooLargeForAllNodes(self):
         """
@@ -269,50 +271,39 @@ class ClusterScalerTest(ToilTest):
         super(ClusterScalerTest, self).setUp()
         self.config = Config()
         self.config.targetTime = 1800
-        self.config.nodeTypes = ['r3.8xlarge', 'c4.8xlarge:0.6']
-        # Set up a stub provisioner with some nodeTypes and nodeShapes.
-        try:
-            # In Python 3 we can use a SimpleNamespace as a mock provisioner
-            self.provisioner = types.SimpleNamespace()
-        except:
-            # In Python 2 we should just be able to tack fields onto an object.
-            # But this has been known to produce:
-            # AttributeError: 'newobject' object has no attribute 'nodeTypes'
-            # So we use an Argparse Namespace instead.
-            import argparse
-            self.provisioner = argparse.Namespace()
-        setattr(self.provisioner, 'nodeTypes', ['r3.8xlarge', 'c4.8xlarge'])
-        setattr(self.provisioner, 'nodeShapes', [r3_8xlarge,
-                                                 c4_8xlarge_preemptable])
-        setattr(self.provisioner, 'setStaticNodes', lambda _, __: None)
-        setattr(self.provisioner, 'retryPredicate', lambda _: False)
+        self.config.nodeTypes = [r3_8xlarge, c4_8xlarge_preemptable]
 
+        # Set up the mock leader
         self.leader = MockBatchSystemAndProvisioner(self.config, 1)
-    
+        # It is also a full mock provisioner, so configure it to be that as well
+        self.provisioner = self.leader
+        # Pretend that Shapes are actually strings we can use for instance type names.
+        self.provisioner.setAutoscaledNodeTypes([({t}, None) for t in self.config.nodeTypes])
+
     @travis_test
     def testRounding(self):
         """
         Test to make sure the ClusterScaler's rounding rounds properly.
         """
-        
+
         # Get a ClusterScaler
         self.config.targetTime = 1
         self.config.betaInertia = 0.0
         self.config.maxNodes = [2, 3]
         scaler = ClusterScaler(self.provisioner, self.leader, self.config)
-        
+
         # Exact integers round to themselves
         self.assertEqual(scaler._round(0.0), 0)
         self.assertEqual(scaler._round(1.0), 1)
         self.assertEqual(scaler._round(-1.0), -1)
         self.assertEqual(scaler._round(123456789101112.13), 123456789101112)
-        
+
         # Decimals other than X.5 round to the side they are closer to
         self.assertEqual(scaler._round(1E-10), 0)
         self.assertEqual(scaler._round(0.5 + 1E-15), 1)
         self.assertEqual(scaler._round(-0.9), -1)
         self.assertEqual(scaler._round(-0.4), 0)
-        
+
         # Decimals at exactly X.5 round away from 0
         self.assertEqual(scaler._round(0.5), 1)
         self.assertEqual(scaler._round(-0.5), -1)
@@ -321,7 +312,7 @@ class ClusterScalerTest(ToilTest):
         self.assertEqual(scaler._round(15.5), 16)
         self.assertEqual(scaler._round(-15.5), -16)
         self.assertEqual(scaler._round(123456789101112.5), 123456789101113)
-    
+
     @travis_test
     def testMaxNodes(self):
         """
@@ -345,7 +336,7 @@ class ClusterScalerTest(ToilTest):
         estimatedNodeCounts = scaler.getEstimatedNodeCounts(jobShapes, defaultdict(int))
         self.assertEqual(estimatedNodeCounts[r3_8xlarge], 2)
         self.assertEqual(estimatedNodeCounts[c4_8xlarge_preemptable], 3)
-    
+
     @travis_test
     def testMinNodes(self):
         """
@@ -358,7 +349,7 @@ class ClusterScalerTest(ToilTest):
         estimatedNodeCounts = scaler.getEstimatedNodeCounts(jobShapes, defaultdict(int))
         self.assertEqual(estimatedNodeCounts[r3_8xlarge], 2)
         self.assertEqual(estimatedNodeCounts[c4_8xlarge_preemptable], 3)
-    
+
     @travis_test
     def testPreemptableDeficitResponse(self):
         """
@@ -376,15 +367,13 @@ class ClusterScalerTest(ToilTest):
         # that we can have preemptable and non-preemptable nodes of
         # the same type. That is the only situation where
         # preemptableCompensation applies.
-        self.config.nodeTypes = ['c4.8xlarge:0.6', 'c4.8xlarge']
-        self.provisioner.nodeTypes = ['c4.8xlarge', 'c4.8xlarge']
-        self.provisioner.nodeShapes = [c4_8xlarge_preemptable,
-                                       c4_8xlarge]
-
+        self.config.nodeTypes = [c4_8xlarge_preemptable, c4_8xlarge]
+        self.provisioner.setAutoscaledNodeTypes([({t}, None) for t in self.config.nodeTypes])
+        
         scaler = ClusterScaler(self.provisioner, self.leader, self.config)
         # Simulate a situation where a previous run caused a
         # "deficit" of 5 preemptable nodes (e.g. a spot bid was lost)
-        scaler.preemptableNodeDeficit['c4.8xlarge'] = 5
+        scaler.preemptableNodeDeficit[c4_8xlarge] = 5
         # Add a bunch of preemptable jobs (so the bin-packing
         # estimate for the non-preemptable node should still be 0)
         jobShapes = [Shape(wallTime=3600,
@@ -396,8 +385,8 @@ class ClusterScalerTest(ToilTest):
         # We don't care about the estimated size of the preemptable
         # nodes. All we want to know is if we responded to the deficit
         # properly: 0.5 * 5 (preemptableCompensation * the deficit) = 3 (rounded up).
-        self.assertEqual(estimatedNodeCounts[self.provisioner.nodeShapes[1]], 3)
-    
+        self.assertEqual(estimatedNodeCounts[self.provisioner.node_shapes_for_testing[1]], 3)
+
     @travis_test
     def testPreemptableDeficitIsSet(self):
         """
@@ -415,22 +404,20 @@ class ClusterScalerTest(ToilTest):
         # that we can have preemptable and non-preemptable nodes of
         # the same type. That is the only situation where
         # preemptableCompensation applies.
-        self.config.nodeTypes = ['c4.8xlarge:0.6', 'c4.8xlarge']
-        self.provisioner.nodeTypes = ['c4.8xlarge', 'c4.8xlarge']
-        self.provisioner.nodeShapes = [c4_8xlarge_preemptable,
-                                       c4_8xlarge]
+        self.config.nodeTypes = [c4_8xlarge_preemptable, c4_8xlarge]
+        self.provisioner.setAutoscaledNodeTypes([({t}, None) for t in self.config.nodeTypes])
         scaler = ClusterScaler(self.provisioner, self.leader, self.config)
         estimatedNodeCounts = {c4_8xlarge_preemptable: 5, c4_8xlarge: 0}
         scaler.updateClusterSize(estimatedNodeCounts)
-        self.assertEqual(scaler.preemptableNodeDeficit['c4.8xlarge'], 2)
+        self.assertEqual(scaler.preemptableNodeDeficit[c4_8xlarge_preemptable], 2)
         self.provisioner.addNodes.assert_called_once()
 
         # OK, now pretend this is a while later, and actually launched
         # the nodes properly. The deficit should disappear
         self.provisioner.addNodes = MagicMock(return_value=5)
         scaler.updateClusterSize(estimatedNodeCounts)
-        self.assertEqual(scaler.preemptableNodeDeficit['c4.8xlarge'], 0)
-    
+        self.assertEqual(scaler.preemptableNodeDeficit[c4_8xlarge], 0)
+
     @travis_test
     def testNoLaunchingIfDeltaAlreadyMet(self):
         """
@@ -438,14 +425,13 @@ class ClusterScalerTest(ToilTest):
         the delta was able to be met by unignoring nodes.
         """
         # We have only one node type for simplicity
-        self.provisioner.nodeTypes = ['c4.8xlarge']
-        self.provisioner.nodeShapes = [c4_8xlarge]
+        self.provisioner.setAutoscaledNodeTypes([({c4_8xlarge}, None)])
         scaler = ClusterScaler(self.provisioner, self.leader, self.config)
         # Pretend there is one ignored worker in the cluster
         self.provisioner.getProvisionedWorkers = MagicMock(
             return_value=[Node('127.0.0.1', '127.0.0.1', 'testNode',
                                datetime.datetime.now().isoformat(),
-                               nodeType='c4.8xlarge', preemptable=True)])
+                               nodeType=c4_8xlarge, preemptable=True)])
         scaler.ignoredNodes.add('127.0.0.1')
         # Exercise the updateClusterSize logic
         self.provisioner.addNodes = MagicMock()
@@ -455,7 +441,7 @@ class ClusterScalerTest(ToilTest):
         self.assertEqual(len(scaler.ignoredNodes), 0,
                          "The scaler didn't unignore an ignored node when "
                          "scaling up")
-    
+
     @travis_test
     def testBetaInertia(self):
         # This is really high, but makes things easy to calculate.
@@ -482,6 +468,7 @@ class ScalerThreadTest(ToilTest):
         # number of worker nodes used.
 
         mock = MockBatchSystemAndProvisioner(config, secondsPerJob=2.0)
+        mock.setAutoscaledNodeTypes([({t}, None) for t in config.nodeTypes])
         mock.start()
         clusterScaler = ScalerThread(mock, mock, config)
         clusterScaler.start()
@@ -592,6 +579,7 @@ class ScalerThreadTest(ToilTest):
         config.scaleInterval = 3
 
         mock = MockBatchSystemAndProvisioner(config, secondsPerJob=2.0)
+        mock.setAutoscaledNodeTypes([({t}, None) for t in config.nodeTypes])
         clusterScaler = ScalerThread(mock, mock, config)
         clusterScaler.start()
         mock.start()
@@ -662,16 +650,13 @@ class MockBatchSystemAndProvisioner(AbstractScalableBatchSystem, AbstractProvisi
     Mimics a job batcher, provisioner and scalable batch system
     """
     def __init__(self, config, secondsPerJob):
-        super(MockBatchSystemAndProvisioner, self).__init__('clusterName')
+        super(MockBatchSystemAndProvisioner, self).__init__(clusterName='clusterName', clusterType='mesos')
         # To mimic parallel preemptable and non-preemptable queues
         # for jobs we create two parallel instances of the following class
         self.config = config
         self.secondsPerJob = secondsPerJob
         self.provisioner = self
         self.batchSystem = self
-        self.nodeTypes = config.nodeTypes
-        self.nodeShapes = self.nodeTypes
-        self.nodeShapes.sort()
         self.jobQueue = Queue()
         self.updatedJobsQueue = Queue()
         self.jobBatchSystemIDToIssuedJob = {}
@@ -679,10 +664,8 @@ class MockBatchSystemAndProvisioner(AbstractScalableBatchSystem, AbstractProvisi
         self.totalWorkerTime = 0.0  # Total time spent in worker threads
         self.toilMetrics = None
         self.nodesToWorker = {}  # Map from Node to instances of the Worker class
-        self.workers = {nodeShape: [] for nodeShape in
-                        self.nodeShapes}  # Instances of the Worker class
-        self.maxWorkers = {nodeShape: 0 for nodeShape in
-                           self.nodeShapes}  # Maximum number of workers
+        self.workers = defaultdict(list)  # Instances of the Worker class, by node shape
+        self.maxWorkers = defaultdict(int)  # Maximum number of workers, by node shape
         self.running = False
         self.leaderThread = Thread(target=self._leaderFn)
 
@@ -711,15 +694,26 @@ class MockBatchSystemAndProvisioner(AbstractScalableBatchSystem, AbstractProvisi
     def unignoreNode(self, nodeAddress):
         pass
 
+    def supportedClusterTypes(self):
+        return {'mesos'}
+
+    def createClusterSettings(self):
+        pass
+
+    def readClusterSettings(self):
+        pass
+
     @contextmanager
     def nodeFiltering(self, filter):
-        nodes = self.getProvisionedWorkers(preemptable=True,
-                                           nodeType=None) + self.getProvisionedWorkers(
-            preemptable=False, nodeType=None)
+        nodes = self.getProvisionedWorkers()
         yield nodes
 
     # AbstractProvisioner methods
-    def getProvisionedWorkers(self, nodeType=None, preemptable=None):
+    def setAutoscaledNodeTypes(self, node_types: List[Tuple[Set[Shape], Optional[float]]]):
+        self.node_shapes_for_testing = sorted([it for t in node_types for it in t[0]])
+        super().setAutoscaledNodeTypes(node_types)
+
+    def getProvisionedWorkers(self, instance_type=None, preemptable=None):
         """
         Returns a list of Node objects, each representing a worker node in the cluster
 
@@ -727,10 +721,14 @@ class MockBatchSystemAndProvisioner(AbstractScalableBatchSystem, AbstractProvisi
         :return: list of Node
         """
         nodesToWorker = self.nodesToWorker
-        if nodeType:
-            return [node for node in nodesToWorker if node.nodeType == nodeType]
+        results = []
+        if instance_type:
+            results = [node for node in nodesToWorker if node.nodeType == instance_type]
         else:
-            return list(nodesToWorker.keys())
+            results = list(nodesToWorker.keys())
+        if preemptable is not None:
+            results = [node for node in results if node.preemptable == preemptable]
+        return results
 
     def terminateNodes(self, nodes):
         self._removeNodes(nodes)
@@ -775,12 +773,14 @@ class MockBatchSystemAndProvisioner(AbstractScalableBatchSystem, AbstractProvisi
         return nodes
 
     # AbstractProvisioner functionality
-    def addNodes(self, nodeType, numNodes, preemptable):
+    def addNodes(self, nodeTypes: Set[str], numNodes, preemptable):
+        nodeType = next(iter(nodeTypes))
         self._addNodes(numNodes=numNodes, nodeType=nodeType, preemptable=preemptable)
         return self.getNumberOfNodes(nodeType=nodeType, preemptable=preemptable)
 
     def getNodeShape(self, nodeType, preemptable=False):
         # Assume node shapes and node types are the same thing for testing
+        # TODO: this isn't really allowed by the type requirements on AbstractProvisioner
         return nodeType
 
     def getWorkersInCluster(self, nodeShape):
@@ -808,7 +808,7 @@ class MockBatchSystemAndProvisioner(AbstractScalableBatchSystem, AbstractProvisi
                 del self.jobBatchSystemIDToIssuedJob[updatedJobID]
             time.sleep(0.1)
 
-    def _addNodes(self, numNodes, nodeType, preemptable=False):
+    def _addNodes(self, numNodes, nodeType: str, preemptable=False):
         nodeShape = self.getNodeShape(nodeType=nodeType, preemptable=preemptable)
 
         class Worker(object):
