@@ -71,13 +71,27 @@ class hidden:
 
         def setUp(self):
             super(hidden.AbstractFileStoreTest, self).setUp()
-            testDir = self._createTempDir()
+            self.work_dir = self._createTempDir()
             self.options = Job.Runner.getDefaultOptions(self._getTestJobStore())
             self.options.logLevel = 'DEBUG'
             self.options.realTimeLogging = True
-            self.options.workDir = testDir
+            self.options.workDir = self.work_dir
             self.options.clean = 'always'
-            self.options.logFile = os.path.join(testDir, 'logFile')
+            self.options.logFile = os.path.join(self.work_dir, 'logFile')
+
+            self.tmp_dir = self._createTempDir()
+
+        def create_file(self, content, executable=False):
+            file_path = f'{self.tmp_dir}/{uuid4()}'
+
+            with open(file_path, 'w') as f:
+                f.write(content)
+
+            if executable:
+                # Add file owner execute permissions
+                os.chmod(file_path, os.stat(file_path).st_mode | stat.S_IXUSR)
+
+            return file_path
 
         @staticmethod
         def _uselessFunc(job):
@@ -118,6 +132,7 @@ class hidden:
                     super().__init__()
                     self.match = match
                     self.seen = False
+
                 def emit(self, record):
                     if self.match in record.getMessage():
                         self.seen = True
@@ -140,11 +155,11 @@ class hidden:
 
         @staticmethod
         def _accessAndFail(job):
-            with job.fileStore.writeGlobalFileStream() as (stream, fileID):
-                stream.write('Cats'.encode('utf-8'))
+            with job.fileStore.writeGlobalFileStream() as (writable, file_id):
+                writable.write(b'Cats')
             localPath = os.path.join(job.fileStore.getLocalTempDir(), 'cats.txt')
-            job.fileStore.readGlobalFile(fileID, localPath)
-            with job.fileStore.readGlobalFileStream(fileID) as stream2:
+            job.fileStore.readGlobalFile(file_id, localPath)
+            with job.fileStore.readGlobalFileStream(file_id) as readable:
                 pass
             raise RuntimeError("I do not like this file")
 
@@ -241,9 +256,14 @@ class hidden:
             are downloaded again. This function checks that a written executable file
             maintains its executability after being read.
             """
-            for executable in True,False:
-                A = Job.wrapJobFn(self._testWriteReadGlobalFilePermissions, executable=executable)
-                Job.Runner.startToil(A, self.options)
+            for executable in True, False:
+                for disable_caching in True, False:
+                    with self.subTest(f'Testing readwrite file permissions\n'
+                                      f'[executable: {executable}]\n'
+                                      f'[disable_caching: {disable_caching}]\n'):
+                        self.options.disableCaching = disable_caching
+                        read_write_job = Job.wrapJobFn(self._testWriteReadGlobalFilePermissions, executable=executable)
+                        Job.Runner.startToil(read_write_job, self.options)
 
         @staticmethod
         def _testWriteReadGlobalFilePermissions(job, executable):
@@ -258,29 +278,28 @@ class hidden:
             initialPermissions = os.stat(srcFile).st_mode & stat.S_IXUSR
             fileID = job.fileStore.writeGlobalFile(srcFile)
 
-            for mutable in True,False:
-                for symlink in True,False:
+            for mutable in True, False:
+                for symlink in True, False:
                     dstFile = job.fileStore.getLocalTempFileName()
                     job.fileStore.readGlobalFile(fileID, userPath=dstFile, mutable=mutable, symlink=symlink)
                     # Current file owner execute permissions
                     currentPermissions = os.stat(dstFile).st_mode & stat.S_IXUSR
-
-                    assert initialPermissions == currentPermissions
+                    assert initialPermissions == currentPermissions, f'{initialPermissions} != {currentPermissions}'
 
         def testWriteExportFileCompatibility(self):
             """
             Ensures that files created in a job preserve their executable permissions
             when they are exported from the leader.
             """
-            for executable in True,False:
-                A = Job.wrapJobFn(self._testWriteExportFileCompatibility, executable=executable)
+            for executable in True, False:
+                export_file_job = Job.wrapJobFn(self._testWriteExportFileCompatibility, executable=executable)
                 with Toil(self.options) as toil:
-                    initialPermissions, fileID = toil.start(A)
+                    initialPermissions, fileID = toil.start(export_file_job)
                     dstFile = os.path.join(self._createTempDir(), str(uuid4()))
                     toil.exportFile(fileID, 'file://' + dstFile)
                     currentPermissions = os.stat(dstFile).st_mode & stat.S_IXUSR
 
-                    assert initialPermissions == currentPermissions
+                    assert initialPermissions == currentPermissions, f'{initialPermissions} != {currentPermissions}'
 
         @staticmethod
         def _testWriteExportFileCompatibility(job, executable):
@@ -299,30 +318,43 @@ class hidden:
             when they are read by the fileStore.
             """
             with Toil(self.options) as toil:
-                workDir = self._createTempDir()
                 for executable in True, False:
-                    srcFile = os.path.join(workDir, str(uuid4()))
-                    with open(srcFile, 'w') as f:
-                        f.write('Hello')
-                    if executable:
-                        os.chmod(srcFile, os.stat(srcFile).st_mode | stat.S_IXUSR)
-                    initialPermissions = os.stat(srcFile).st_mode & stat.S_IXUSR
-                    jobStoreFileID = toil.importFile('file://' + srcFile)
-                    for mutable in True,False:
+                    file_path = self.create_file(content='Hello', executable=executable)
+                    initial_permissions = os.stat(file_path).st_mode & stat.S_IXUSR
+                    file_id = toil.importFile(f'file://{file_path}')
+                    for mutable in True, False:
                         for symlink in True, False:
                             with self.subTest(f'Now testing readGlobalFileWith: mutable={mutable} symlink={symlink}'):
-                                dstFile = os.path.join(workDir, str(uuid4()))
-                                A = Job.wrapJobFn(_testImportReadFileCompatibility, fileID=jobStoreFileID, dstFile=dstFile,
-                                                    initialPermissions=initialPermissions, mutable=mutable, symlink=symlink)
+                                A = Job.wrapJobFn(self._testImportReadFileCompatibility,
+                                                  fileID=file_id,
+                                                  initialPermissions=initial_permissions,
+                                                  mutable=mutable,
+                                                  symlink=symlink)
                                 toil.start(A)
 
         @staticmethod
-        def _testImportReadFileCompatibility(job, fileID, dstFile, initialPermissions, mutable, symlink):
+        def _testImportReadFileCompatibility(job, fileID, initialPermissions, mutable, symlink):
             dstFile = job.fileStore.readGlobalFile(fileID, mutable=mutable, symlink=symlink)
             currentPermissions = os.stat(dstFile).st_mode & stat.S_IXUSR
 
             assert initialPermissions == currentPermissions
 
+        def testReadWriteFileStreamTextMode(self):
+            """
+            Checks if text mode is compatible with file streams.
+            """
+            with Toil(self.options) as toil:
+                A = Job.wrapJobFn(self._testReadWriteFileStreamTextMode)
+                toil.start(A)
+
+        @staticmethod
+        def _testReadWriteFileStreamTextMode(job):
+            with job.fileStore.writeGlobalFileStream(encoding='utf-8') as (stream, fileID):
+                stream.write('foo')
+            job.fileStore.readGlobalFileStream(fileID)
+            with job.fileStore.readGlobalFileStream(fileID, encoding='utf-8') as stream2:
+                assert 'foo' == stream2.read()
+
         @staticmethod
         def _writeFileToJobStore(job, isLocalFile, nonLocalDir=None, fileMB=1):
             """
@@ -467,12 +499,12 @@ class hidden:
                 B = Job.wrapJobFn(self._sleepy, timeToSleep=1)
                 C = Job.wrapJobFn(self._writeFileToJobStoreWithAsserts, isLocalFile=True,
                                   fileMB=file2MB)
-                D = Job.wrapJobFn(self._adjustCacheLimit, newTotalMB=50, disk='0M')
-                E = Job.wrapJobFn(self._uselessFunc, disk=''.join([str(diskRequestMB), 'M']))
+                D = Job.wrapJobFn(self._adjustCacheLimit, newTotalMB=50, disk='0Mi')
+                E = Job.wrapJobFn(self._uselessFunc, disk=''.join([str(diskRequestMB), 'Mi']))
                 # Set it to > 2GB such that the cleanup jobs don't die in the non-fail cases
-                F = Job.wrapJobFn(self._adjustCacheLimit, newTotalMB=5000, disk='10M')
+                F = Job.wrapJobFn(self._adjustCacheLimit, newTotalMB=5000, disk='10Mi')
                 G = Job.wrapJobFn(self._probeJobReqs, sigmaJob=100, cached=expectedResult,
-                                  disk='100M')
+                                  disk='100Mi')
                 A.addChild(B)
                 B.addChild(C)
                 C.addChild(D)
@@ -798,12 +830,12 @@ class hidden:
             A = Job.wrapJobFn(self._writeFileToJobStoreWithAsserts, isLocalFile=cacheHit,
                               nonLocalDir=workdir,
                               fileMB=256)
-            B = Job.wrapJobFn(self._probeJobReqs, sigmaJob=100, disk='100M')
+            B = Job.wrapJobFn(self._probeJobReqs, sigmaJob=100, disk='100Mi')
             jobs = {}
             for i in range(0, 10):
                 jobs[i] = Job.wrapJobFn(self._multipleFileReader, diskMB=1024, fsID=A.rv(),
-                                        maxWriteFile=os.path.abspath(x.name), disk='1G',
-                                        memory='10M', cores=1)
+                                        maxWriteFile=os.path.abspath(x.name), disk='1Gi',
+                                        memory='10Mi', cores=1)
                 A.addChild(jobs[i])
                 jobs[i].addChild(B)
             Job.Runner.startToil(A, self.options)
@@ -894,7 +926,7 @@ class hidden:
                               jobDisk=2 * 1024 * 1024 * 1024,
                               initialCachedSize=0,
                               nonLocalDir=workdir,
-                              disk='2G')
+                              disk='2Gi')
             Job.Runner.startToil(F, self.options)
 
         @slow
@@ -912,7 +944,7 @@ class hidden:
                               jobDisk=2 * 1024 * 1024 * 1024,
                               initialCachedSize=0,
                               nonLocalDir=workdir,
-                              numIters=30, disk='2G')
+                              numIters=30, disk='2Gi')
             Job.Runner.startToil(F, self.options)
 
         @staticmethod
@@ -1079,7 +1111,7 @@ class hidden:
             F = Job.wrapJobFn(self._controlledFailTestFn, jobDisk=jobDiskBytes,
                               testDir=workdir,
                               disk=jobDiskBytes)
-            G = Job.wrapJobFn(self._probeJobReqs, sigmaJob=100, disk='100M')
+            G = Job.wrapJobFn(self._probeJobReqs, sigmaJob=100, disk='100Mi')
             F.addChild(G)
             Job.Runner.startToil(F, self.options)
 
