@@ -355,6 +355,15 @@ class WdlLanguageSpecWorkflowsTest(WdlWorkflowsTest):
         expected_result = '[["A Whale of a Tale."],["true"],["11"]]'
         self.check_function('type_pair', cases=['with_files'], expected_result=expected_result)
 
+    def test_v1_declaration(self):
+        """
+        Basic declaration example modified from the WDL 1.0 spec:
+
+        https://github.com/openwdl/wdl/blob/main/versions/1.0/SPEC.md#declarations
+        """
+        expected_result = 'Hello, x!; Hello, y!'
+        self.check_function('v1_spec', cases=['declaration'], expected_result=expected_result)
+
 
 class WdlStandardLibraryWorkflowsTest(WdlWorkflowsTest):
     """
@@ -475,6 +484,23 @@ class WdlStandardLibraryWorkflowsTest(WdlWorkflowsTest):
                                             '{"left":2,"right":"a"},{"left":2,"right":"b"},'
                                             '{"left":3,"right":"a"},{"left":3,"right":"b"}]')
 
+    def test_as_pairs(self):
+        self.check_function('as_pairs', cases=['as_input'],
+                            expected_result='[{"left":"a","right":1},{"left":"b","right":2},{"left":"c","right":3}]')
+
+    def test_as_map(self):
+        self.check_function('as_map', cases=['as_input'], expected_result='{"a":1,"b":2,"c":3}')
+
+    def test_keys(self):
+        self.check_function('keys', cases=['as_input'], expected_result='["a","b","c"]')
+
+    def test_collect_by_key(self):
+        # NOTE: this result is expected according to the spec but differs from Cromwell.
+        self.check_function('collect_by_key', cases=['as_input'], expected_result='{"a":[1,3],"b":[2]}')
+
+    def test_flatten(self):
+        self.check_function('flatten', cases=['as_input'], expected_result='[1,2,3,1,21,22]')
+
 
 if __name__ == "__main__":
     unittest.main()
