@@ -12,35 +12,20 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 """Debug tool for copying files contained in a toil jobStore."""
-import fnmatch
 import logging
 import os.path
 
 from toil.common import Config, Toil, parser_with_common_options
+from toil.jobStores.abstractJobStore import AbstractJobStore
 from toil.statsAndLogging import set_logging_from_options
+from toil.lib.resources import glob
+from toil.lib.expando import Expando
+from typing import Optional
 
 logger = logging.getLogger(__name__)
 
 
-def recursiveGlob(directoryname, glob_pattern):
-    '''
-    Walks through a directory and its subdirectories looking for files matching
-    the glob_pattern and returns a list=[].
-
-    :param directoryname: Any accessible folder name on the filesystem.
-    :param glob_pattern: A string like "*.txt", which would find all text files.
-    :return: A list=[] of absolute filepaths matching the glob pattern.
-    '''
-    directoryname = os.path.abspath(directoryname)
-    matches = []
-    for root, dirnames, filenames in os.walk(directoryname):
-        for filename in fnmatch.filter(filenames, glob_pattern):
-            absolute_filepath = os.path.join(root, filename)
-            matches.append(absolute_filepath)
-    return matches
-
-
-def fetchJobStoreFiles(jobStore, options):
+def fetchJobStoreFiles(jobStore: AbstractJobStore, options: Expando) -> None:
     """
     Takes a list of file names as glob patterns, searches for these within a
     given directory, and attempts to take all of the files found and copy them
@@ -53,8 +38,8 @@ def fetchJobStoreFiles(jobStore, options):
     :param options.jobStore: The path to the jobStore directory.
     """
     for jobStoreFile in options.fetch:
-        jobStoreHits = recursiveGlob(directoryname=options.jobStore,
-                                     glob_pattern=jobStoreFile)
+        jobStoreHits = glob(directoryname=options.jobStore,
+                            glob_pattern=jobStoreFile)
         for jobStoreFileID in jobStoreHits:
             logger.debug(f"Copying job store file: {jobStoreFileID} to {options.localFilePath[0]}")
             jobStore.readFile(jobStoreFileID,
@@ -63,7 +48,7 @@ def fetchJobStoreFiles(jobStore, options):
                               symlink=options.useSymlinks)
 
 
-def printContentsOfJobStore(jobStorePath, nameOfJob=None):
+def printContentsOfJobStore(jobStorePath: str, nameOfJob: Optional[str] = None) -> None:
     """
     Fetch a list of all files contained in the jobStore directory input if
     nameOfJob is not declared, otherwise it only prints out the names of files
@@ -77,14 +62,14 @@ def printContentsOfJobStore(jobStorePath, nameOfJob=None):
     """
 
     if nameOfJob:
-        glob = "*" + nameOfJob + "*"
+        glob_pattern = "*" + nameOfJob + "*"
         logFile = nameOfJob + "_fileset.txt"
     else:
-        glob = "*"
+        glob_pattern = "*"
         logFile = "jobstore_files.txt"
         nameOfJob = ""
 
-    list_of_files = recursiveGlob(directoryname=jobStorePath, glob_pattern=glob)
+    list_of_files = glob(directoryname=jobStorePath, glob_pattern=glob_pattern)
     if os.path.exists(logFile):
         os.remove(logFile)
     for gfile in sorted(list_of_files):
@@ -95,7 +80,7 @@ def printContentsOfJobStore(jobStorePath, nameOfJob=None):
                 f.write("\n")
 
 
-def main():
+def main() -> None:
     parser = parser_with_common_options(jobstore_option=True)
     parser.add_argument("--localFilePath",
                         nargs=1,
