@@ -17,13 +17,14 @@ import os
 
 from toil import applianceSelf
 from toil.common import parser_with_common_options
-from toil.provisioners import check_valid_node_types, cluster_factory
+from toil.provisioners import parse_node_types, check_valid_node_types, cluster_factory
 from toil.statsAndLogging import set_logging_from_options
+from typing import List, Dict, Union, Tuple
 
 logger = logging.getLogger(__name__)
 
 
-def create_tags_dict(tags: list) -> dict:
+def create_tags_dict(tags: List[str]) -> Dict[str, str]:
     tags_dict = dict()
     for tag in tags:
         key, value = tag.split('=')
@@ -31,8 +32,11 @@ def create_tags_dict(tags: list) -> dict:
     return tags_dict
 
 
-def main():
+def main() -> None:
     parser = parser_with_common_options(provisioner_options=True, jobstore_option=False)
+    parser.add_argument("-T", "--clusterType", dest="clusterType",
+                        choices=['mesos', 'kubernetes'], default='mesos',
+                        help="Cluster scheduler to use.")
     parser.add_argument("--leaderNodeType", dest="leaderNodeType", required=True,
                         help="Non-preemptable node type to use for the cluster leader.")
     parser.add_argument("--keyPairName", dest='keyPairName',
@@ -56,24 +60,34 @@ def main():
                              "      \"Owner\": IAM username\n"
                              " }. ")
     parser.add_argument("--vpcSubnet",
-                        help="VPC subnet ID to launch cluster in. Uses default subnet if not "
-                        "specified. This subnet needs to have auto assign IPs turned on.")
+                        help="VPC subnet ID to launch cluster leader in. Uses default subnet "
+                        "if not specified. This subnet needs to have auto assign IPs turned on.")
     parser.add_argument("--nodeTypes", dest='nodeTypes', default=None, type=str,
-                        help="Comma-separated list of node types to create while launching the "
-                             "leader. The syntax for each node type depends on the provisioner "
-                             "used. For the aws provisioner this is the name of an EC2 instance "
-                             "type followed by a colon and the price in dollar to bid for a spot "
-                             "instance, for example 'c3.8xlarge:0.42'. Must also provide the "
-                             "--workers argument to specify how many workers of each node type "
-                             "to create.")
+                        help="Specifies a list of comma-separated node types, each of which is "
+                             "composed of slash-separated instance types, and an optional spot "
+                             "bid set off by a colon, making the node type preemptable. Instance "
+                             "types may appear in multiple node types, and the same node type "
+                             "may appear as both preemptable and non-preemptable.\n"
+                             "Valid argument specifying two node types:\n"
+                             "\tc5.4xlarge/c5a.4xlarge:0.42,t2.large\n"
+                             "Node types:\n"
+                             "\tc5.4xlarge/c5a.4xlarge:0.42 and t2.large\n"
+                             "Instance types:\n"
+                             "\tc5.4xlarge, c5a.4xlarge, and t2.large\n"
+                             "Semantics:\n"
+                             "\tBid $0.42/hour for either c5.4xlarge or c5a.4xlarge instances,\n"
+                             "\ttreated interchangeably, while they are available at that price,\n"
+                             "\tand buy t2.large instances at full price\n"
+                             "Must also provide the --workers argument to specify how many "
+                             "workers of each node type to create.")
     parser.add_argument("-w", "--workers", dest='workers', default=None, type=str,
-                        help="Comma-separated list of the number of workers of each node type to "
-                             "launch alongside the leader when the cluster is created. This can be "
-                             "useful if running toil without auto-scaling but with need of more "
-                             "hardware support")
+                        help="Comma-separated list of the ranges of numbers of workers of each "
+                             "node type to launch, such as '0-2,5,1-3'. If a range is given, "
+                             "workers will automatically be launched and terminated by the cluster "
+                             "to auto-scale to the workload.")
     parser.add_argument("--leaderStorage", dest='leaderStorage', type=int, default=50,
                         help="Specify the size (in gigabytes) of the root volume for the leader "
-                             "instance.  This is an EBS volume.")
+                             "instance. This is an EBS volume.")
     parser.add_argument("--nodeStorage", dest='nodeStorage', type=int, default=50,
                         help="Specify the size (in gigabytes) of the root volume for any worker "
                              "instances created when using the -w flag. This is an EBS volume.")
@@ -94,13 +108,38 @@ def main():
     set_logging_from_options(options)
     tags = create_tags_dict(options.tags) if options.tags else dict()
 
-    worker_node_types = options.nodeTypes.split(',') if options.nodeTypes else []
-    worker_quantities = options.workers.split(',') if options.workers else []
-    check_valid_node_types(options.provisioner, worker_node_types + [options.leaderNodeType])
+    # Get worker node types
+    worker_node_types = parse_node_types(options.nodeTypes)
+    check_valid_node_types(options.provisioner, worker_node_types + [({options.leaderNodeType}, None)])
+
+    # Holds string ranges, like "5", or "3-10"
+    worker_node_ranges = options.workers.split(',') if options.workers else []
 
     # checks the validity of TOIL_APPLIANCE_SELF before proceeding
     applianceSelf(forceDockerAppliance=options.forceDockerAppliance)
 
+    # This holds either ints to launch static nodes, or tuples of ints
+    # specifying ranges to launch managed auto-scaling nodes, for each type.
+    nodeCounts: List[Union[int, Tuple[int, int]]] = []
+
+    if ((worker_node_types != [] or worker_node_ranges != []) and not
+        (worker_node_types != [] and worker_node_ranges != [])):
+        raise RuntimeError("The --nodeTypes option requires --workers, and visa versa.")
+    if worker_node_types and worker_node_ranges:
+            if not len(worker_node_types) == len(worker_node_ranges):
+                raise RuntimeError("List of worker count ranges must be the same length as the list of node types.")
+
+            for spec in worker_node_ranges:
+                if '-' in spec:
+                    # Provision via autoscaling
+                    parts = spec.split('-')
+                    if len(parts) != 2:
+                        raise RuntimeError("Unacceptable range: " + spec)
+                    nodeCounts.append((int(parts[0]), int(parts[1])))
+                else:
+                    # Provision fixed nodes
+                    nodeCounts.append(int(spec))
+
     owner = options.owner or options.keyPairName or 'toil'
 
     # Check to see if the user specified a zone. If not, see if one is stored in an environment variable.
@@ -110,14 +149,11 @@ def main():
         raise RuntimeError(f'Please provide a value for --zone or set a default in the '
                            f'TOIL_{options.provisioner.upper()}_ZONE environment variable.')
 
-    if (options.nodeTypes or options.workers) and not (options.nodeTypes and options.workers):
-        raise RuntimeError("The --nodeTypes and --workers options must be specified together.")
-
-    if not len(worker_node_types) == len(worker_quantities):
-        raise RuntimeError("List of node types must be the same length as the list of workers.")
+    logger.info('Creating cluster %s...', options.clusterName)
 
     cluster = cluster_factory(provisioner=options.provisioner,
                               clusterName=options.clusterName,
+                              clusterType=options.clusterType,
                               zone=options.zone,
                               nodeStorage=options.nodeStorage)
 
@@ -131,14 +167,45 @@ def main():
                           awsEc2ProfileArn=options.awsEc2ProfileArn,
                           awsEc2ExtraSecurityGroupIds=options.awsEc2ExtraSecurityGroupIds)
 
-    for worker_node_type, num_workers in zip(worker_node_types, worker_quantities):
-        if ':' in worker_node_type:
-            worker_node_type, bid = worker_node_type.split(':', 1)
-            cluster.addNodes(nodeType=worker_node_type,
-                             numNodes=int(num_workers),
-                             preemptable=True,
-                             spotBid=float(bid))
-        else:
-            cluster.addNodes(nodeType=worker_node_type,
-                             numNodes=int(num_workers),
-                             preemptable=False)
+    for typeNum, spec in enumerate(nodeCounts):
+        # For each batch of workers to make
+        wanted = worker_node_types[typeNum]
+
+        if isinstance(spec, int):
+            # Make static nodes
+
+            if spec == 0:
+                # Don't make anything
+                continue
+
+            if wanted[1] is None:
+                # Make non-spot instances
+                cluster.addNodes(nodeTypes=wanted[0], numNodes=spec, preemptable=False)
+            else:
+                # We have a spot bid
+                cluster.addNodes(nodeTypes=wanted[0], numNodes=spec, preemptable=True,
+                                 spotBid=wanted[1])
+
+        elif isinstance(spec, tuple):
+            # Make a range of auto-scaling nodes
+
+            max_count, min_count = spec
+
+            if max_count < min_count:
+                # Flip them around
+                min_count, max_count = max_count, min_count
+
+            if max_count == 0:
+                # Don't want any
+                continue
+
+            if wanted[1] is None:
+                # Make non-spot instances
+                cluster.addManagedNodes(nodeTypes=wanted[0], minNodes=min_count, maxNodes=max_count,
+                                        preemptable=False)
+            else:
+                # Bid at the given price.
+                cluster.addManagedNodes(nodeTypes=wanted[0], minNodes=min_count, maxNodes=max_count,
+                                        preemptable=True, spotBid=wanted[1])
+
+    logger.info('Cluster created successfully.')
