@@ -2,13 +2,14 @@ import os
 import re
 import sys
 import textwrap
+import types
 
 import pkg_resources
 
 from toil.version import version
+from typing import Any, Dict
 
-
-def main():
+def main() -> None:
     modules = loadModules()
 
     if len(sys.argv) < 2 or sys.argv[1] == '--help':
@@ -22,14 +23,25 @@ def main():
 
     try:
         module = modules[cmd]
-        del sys.argv[1]
-        module.main()
     except KeyError:
         sys.stderr.write(f'Unknown option "{cmd}".  Pass --help to display usage information.\n')
         sys.exit(1)
 
+    del sys.argv[1]
+    get_or_die(module, 'main')()
 
-def loadModules():
+def get_or_die(module: types.ModuleType, name: str) -> Any:
+    """
+    Get an object from a module or complain that it is missing.
+    """
+    if hasattr(module, name):
+        return getattr(module, name)
+    else:
+        sys.stderr.write(f'Internal Toil error!\nToil utility module '
+                         f'{module.__name__} is missing required attribute {name}\n')
+        sys.exit(1)
+
+def loadModules() -> Dict[str, types.ModuleType]:
     # noinspection PyUnresolvedReferences
     from toil.utils import (toilClean,
                             toilDebugFile,
@@ -44,21 +56,21 @@ def loadModules():
     return {"-".join([i.lower() for i in re.findall('[A-Z][^A-Z]*', name)]): module for name, module in locals().items()}
 
 
-def printHelp(modules):
+def printHelp(modules: Dict[str, types.ModuleType]) -> None:
     name = os.path.basename(sys.argv[0])
-    descriptions = '\n        '.join(f'{cmd} - {mod.__doc__.strip()}' for cmd, mod in modules.items() if mod)
+    descriptions = '\n        '.join(f'{cmd} - {get_or_die(mod, "__doc__").strip()}' for cmd, mod in modules.items() if mod)
     print(textwrap.dedent(f"""
         Usage: {name} COMMAND ...
                {name} --help
                {name} COMMAND --help
 
         Where COMMAND is one of the following:
-        
+
         {descriptions}
         """[1:]))
 
 
-def printVersion():
+def printVersion() -> None:
     try:
         print(pkg_resources.get_distribution('toil').version)
     except:
