@@ -22,39 +22,47 @@ import random
 import shutil
 import signal
 import socket
+import stat
 import sys
-import tempfile
 import time
 import traceback
+
 from contextlib import contextmanager
+from typing import Any, Callable, Iterator, List, Optional
 
 from toil import logProcessContext
-from toil.common import Toil, safeUnpickleFromStream
+from toil.common import Toil, Config, safeUnpickleFromStream
 from toil.deferred import DeferredFunctionManager
 from toil.fileStores.abstractFileStore import AbstractFileStore
-from toil.job import CheckpointJobDescription, Job
+from toil.job import CheckpointJobDescription, Job, JobDescription
+from toil.jobStores.abstractJobStore import AbstractJobStore
 from toil.lib.expando import MagicExpando
+from toil.lib.io import make_public_dir
 from toil.lib.resources import (get_total_cpu_time,
                                 get_total_cpu_time_and_memory_usage)
 from toil.statsAndLogging import configure_root_logger, set_log_level
 
 try:
-    from toil.cwl.cwltoil import CWL_INTERNAL_JOBS
+    from toil.cwl.cwltoil import (CWL_INTERNAL_JOBS,
+                                  CWL_UNSUPPORTED_REQUIREMENT_EXIT_CODE,
+                                  CWL_UNSUPPORTED_REQUIREMENT_EXCEPTION)
 except ImportError:
     # CWL extra not installed
     CWL_INTERNAL_JOBS = ()
+    CWL_UNSUPPORTED_REQUIREMENT_EXIT_CODE = None
+    CWL_UNSUPPORTED_REQUIREMENT_EXCEPTION = type(None)
 
 logger = logging.getLogger(__name__)
 
-def nextChainable(predecessor, jobStore, config):
+
+def nextChainable(predecessor: JobDescription, jobStore: AbstractJobStore, config: Config) -> Optional[JobDescription]:
     """
     Returns the next chainable job's JobDescription after the given predecessor
     JobDescription, if one exists, or None if the chain must terminate.
-    
-    :param toil.job.JobDescription predecessor: The job to chain from
-    :param toil.jobStores.abstractJobStore.AbstractJobStore jobStore: The JobStore to fetch JobDescriptions from.
-    :param toil.common.Config config: The configuration for the current run.
-    :rtype: toil.job.JobDescription or None
+
+    :param predecessor: The job to chain from
+    :param jobStore: The JobStore to fetch JobDescriptions from.
+    :param config: The configuration for the current run.
     """
     #If no more jobs to run or services not finished, quit
     if len(predecessor.stack) == 0 or len(predecessor.services) > 0 or (isinstance(predecessor, CheckpointJobDescription) and predecessor.checkpoint != None):
@@ -86,7 +94,7 @@ def nextChainable(predecessor, jobStore, config):
 
     # Grab the only job that should be there.
     successorID = next(iter(jobs))
-    
+
     # Load the successor JobDescription
     successor = jobStore.load(successorID)
 
@@ -116,21 +124,22 @@ def nextChainable(predecessor, jobStore, config):
         # Check if job is a checkpoint job and quit if so
         logger.debug("Next job is checkpoint, so finishing")
         return None
-    
+
     # Made it through! This job is chainable.
     return successor
 
-def workerScript(jobStore, config, jobName, jobStoreID, redirectOutputToLogFile=True):
+def workerScript(jobStore: AbstractJobStore, config: Config, jobName: str, jobStoreID: str, redirectOutputToLogFile: bool = True) -> int:
     """
-    Worker process script, runs a job. 
-    
-    :param str jobName: The "job name" (a user friendly name) of the job to be run
-    :param str jobStoreLocator: Specifies the job store to use
-    :param str jobStoreID: The job store ID of the job to be run
-    
+    Worker process script, runs a job.
+
+    :param jobStore: The JobStore to fetch JobDescriptions from.
+    :param config: The configuration for the current run.
+    :param jobName: The "job name" (a user friendly name) of the job to be run
+    :param jobStoreID: The job store ID of the job to be run
+
     :return int: 1 if a job failed, or 0 if all jobs succeeded
     """
-    
+
     configure_root_logger()
     set_log_level(config.logLevel)
 
@@ -139,7 +148,7 @@ def workerScript(jobStore, config, jobName, jobStoreID, redirectOutputToLogFile=
     ##########################################
 
     logFileByteReportLimit = config.maxLogFileSize
-    
+
     if config.badWorker > 0 and random.random() < config.badWorker:
         # We need to kill the process we are currently in, to simulate worker
         # failure. We don't want to just send SIGKILL, because we can't tell
@@ -152,7 +161,7 @@ def workerScript(jobStore, config, jobName, jobStoreID, redirectOutputToLogFile=
         # process won't wait around for its timeout to expire. I think this is
         # better than the old thread-based way where all of Toil would wait
         # around to be killed.
-        
+
         killTarget = os.getpid()
         sleepTime = config.badWorkerFailInterval * random.random()
         if os.fork() == 0:
@@ -171,14 +180,14 @@ def workerScript(jobStore, config, jobName, jobStoreID, redirectOutputToLogFile=
                 pass
             # Exit without doing any of Toil's cleanup
             os._exit(0)
-            
+
         # We don't need to reap the child. Either it kills us, or we finish
         # before it does. Either way, init will have to clean it up for us.
 
     ##########################################
     #Load the environment for the job
     ##########################################
-    
+
     #First load the environment for the job.
     with jobStore.readSharedFileStream("environment.pickle") as fileHandle:
         environment = safeUnpickleFromStream(fileHandle)
@@ -211,14 +220,12 @@ def workerScript(jobStore, config, jobName, jobStoreID, redirectOutputToLogFile=
             if e != '':
                 sys.path.append(e)
 
-    toilWorkflowDir = Toil.getLocalWorkflowDir(config.workflowID, config.workDir)
-
     ##########################################
     #Setup the temporary directories.
     ##########################################
-
     # Dir to put all this worker's temp files in.
-    localWorkerTempDir = tempfile.mkdtemp(dir=toilWorkflowDir)
+    toilWorkflowDir = Toil.getLocalWorkflowDir(config.workflowID, config.workDir)
+    localWorkerTempDir = make_public_dir(in_directory=toilWorkflowDir)
     os.chmod(localWorkerTempDir, 0o755)
 
     ##########################################
@@ -230,10 +237,10 @@ def workerScript(jobStore, config, jobName, jobStoreID, redirectOutputToLogFile=
     #for this process and all children. Consequently, we can't just replace
     #sys.stdout and sys.stderr; we need to mess with the underlying OS-level
     #file descriptors. See <http://stackoverflow.com/a/11632982/402891>
-    
+
     #When we start, standard input is file descriptor 0, standard output is
     #file descriptor 1, and standard error is file descriptor 2.
-    
+
     # Do we even want to redirect output? Let the config make us not do it.
     redirectOutputToLogFile = redirectOutputToLogFile and not config.disableWorkerOutputCapture
 
@@ -246,7 +253,7 @@ def workerScript(jobStore, config, jobName, jobStoreID, redirectOutputToLogFile=
         logger.info("Redirecting logging to %s", tempWorkerLogPath)
         sys.stdout.flush()
         sys.stderr.flush()
-        
+
         # Save the original stdout and stderr (by opening new file descriptors
         # to the same files)
         origStdOut = os.dup(1)
@@ -276,6 +283,7 @@ def workerScript(jobStore, config, jobName, jobStoreID, redirectOutputToLogFile=
     ##########################################
 
     jobAttemptFailed = False
+    failure_exit_code = 1
     statsDict = MagicExpando()
     statsDict.jobs = []
     statsDict.workers.logsToMaster = []
@@ -287,9 +295,9 @@ def workerScript(jobStore, config, jobName, jobStoreID, redirectOutputToLogFile=
         #Put a message at the top of the log, just to make sure it's working.
         logger.info("---TOIL WORKER OUTPUT LOG---")
         sys.stdout.flush()
-        
+
         logProcessContext(config)
-        
+
         ##########################################
         #Connect to the deferred function system
         ##########################################
@@ -298,15 +306,15 @@ def workerScript(jobStore, config, jobName, jobStoreID, redirectOutputToLogFile=
         ##########################################
         #Load the JobDescription
         ##########################################
-        
+
         jobDesc = jobStore.load(jobStoreID)
         listOfJobs[0] = str(jobDesc)
         logger.debug("Parsed job description")
-        
+
         ##########################################
         #Cleanup from any earlier invocation of the job
         ##########################################
-        
+
         if jobDesc.command == None:
             logger.debug("Job description has no body to run.")
             # Cleanup jobs already finished
@@ -315,7 +323,7 @@ def workerScript(jobStore, config, jobName, jobStoreID, redirectOutputToLogFile=
             jobDesc.filterServiceHosts(predicate)
             logger.debug("Cleaned up any references to completed successor jobs")
 
-        # This cleans the old log file which may 
+        # This cleans the old log file which may
         # have been left if the job is being retried after a job failure.
         oldLogFile = jobDesc.logJobStoreFileID
         if oldLogFile != None:
@@ -349,7 +357,7 @@ def workerScript(jobStore, config, jobName, jobStoreID, redirectOutputToLogFile=
         ##########################################
         #Setup the stats, if requested
         ##########################################
-        
+
         if config.stats:
             startClock = get_total_cpu_time()
 
@@ -358,9 +366,9 @@ def workerScript(jobStore, config, jobName, jobStoreID, redirectOutputToLogFile=
             ##########################################
             #Run the job body, if there is one
             ##########################################
-            
+
             logger.info("Working on job %s", jobDesc)
-            
+
             if jobDesc.command is not None:
                 assert jobDesc.command.startswith("_toil ")
                 logger.debug("Got a command to run: %s" % jobDesc.command)
@@ -381,7 +389,7 @@ def workerScript(jobStore, config, jobName, jobStoreID, redirectOutputToLogFile=
                         with fileStore.open(job):
                             # Get the next block function to wait on committing this job
                             blockFn = fileStore.waitForCommit
-                            
+
                             # Run the job, save new successors, and set up
                             # locally (but don't commit) successor
                             # relationships and job completion.
@@ -395,7 +403,7 @@ def workerScript(jobStore, config, jobName, jobStoreID, redirectOutputToLogFile=
 
                 # Accumulate messages from this job & any subsequent chained jobs
                 statsDict.workers.logsToMaster += fileStore.loggingMessages
-                
+
                 logger.info("Completed body for %s", jobDesc)
 
             else:
@@ -404,8 +412,8 @@ def workerScript(jobStore, config, jobName, jobStoreID, redirectOutputToLogFile=
                 #been scheduled after a failure to cleanup
                 logger.debug("No user job to run, so finishing")
                 break
-            
-            if AbstractFileStore._terminateEvent.isSet():
+
+            if AbstractFileStore._terminateEvent.is_set():
                 raise RuntimeError("The termination flag is set")
 
             ##########################################
@@ -414,13 +422,13 @@ def workerScript(jobStore, config, jobName, jobStoreID, redirectOutputToLogFile=
             successor = nextChainable(jobDesc, jobStore, config)
             if successor is None or config.disableChaining:
                 # Can't chain any more jobs. We are going to stop.
-                
+
                 logger.info("Not chaining from job %s", jobDesc)
-                
-                # TODO: Somehow the commit happens even if we don't start it here. 
-                
+
+                # TODO: Somehow the commit happens even if we don't start it here.
+
                 break
-                
+
             logger.info("Chaining from %s to %s", jobDesc, successor)
 
             ##########################################
@@ -434,7 +442,7 @@ def workerScript(jobStore, config, jobName, jobStoreID, redirectOutputToLogFile=
             # Make sure nothing has gone wrong and we can really chain
             assert jobDesc.memory >= successor.memory
             assert jobDesc.cores >= successor.cores
-           
+
             # Save the successor's original ID, so we can clean it (and its
             # body) up after we finish executing it.
             successorID = successor.jobStoreID
@@ -445,7 +453,7 @@ def workerScript(jobStore, config, jobName, jobStoreID, redirectOutputToLogFile=
             # Now we need to become that successor, under the original ID.
             successor.replace(jobDesc)
             jobDesc = successor
-            
+
             # Problem: successor's job body is a file that will be cleaned up
             # when we delete the successor job by ID. We can't just move it. So
             # we need to roll up the deletion of the successor job by ID with
@@ -455,26 +463,26 @@ def workerScript(jobStore, config, jobName, jobStoreID, redirectOutputToLogFile=
             # Clone the now-current JobDescription (which used to be the successor).
             # TODO: Why??? Can we not?
             jobDesc = copy.deepcopy(jobDesc)
-            
+
             # Build a fileStore to update the job and commit the replacement.
             # TODO: can we have a commit operation without an entire FileStore???
             fileStore = AbstractFileStore.createFileStore(jobStore, jobDesc, localWorkerTempDir, blockFn,
                                                           caching=not config.disableCaching)
-                                                          
+
             # Update blockFn to wait for that commit operation.
             blockFn = fileStore.waitForCommit
 
             # This will update the job once the previous job is done updating
-            fileStore.startCommit(jobState=True)            
-            
+            fileStore.startCommit(jobState=True)
+
             # Clone the current job description again, so that further updates
             # to it (such as new successors being added when it runs) occur
             # after the commit process we just kicked off, and aren't committed
             # early or partially.
             jobDesc = copy.deepcopy(jobDesc)
-            
+
             logger.debug("Starting the next job")
-        
+
         ##########################################
         #Finish up the stats
         ##########################################
@@ -487,31 +495,35 @@ def workerScript(jobStore, config, jobName, jobStoreID, redirectOutputToLogFile=
         # log the worker log path here so that if the file is truncated the path can still be found
         if redirectOutputToLogFile:
             logger.info("Worker log can be found at %s. Set --cleanWorkDir to retain this log", localWorkerTempDir)
-        
+
         logger.info("Finished running the chain of jobs on this node, we ran for a total of %f seconds", time.time() - startTime)
-    
+
     ##########################################
     #Trapping where worker goes wrong
     ##########################################
-    except: #Case that something goes wrong in worker
+    except Exception as e: #Case that something goes wrong in worker
         traceback.print_exc()
         logger.error("Exiting the worker because of a failed job on host %s", socket.gethostname())
+        if isinstance(e, CWL_UNSUPPORTED_REQUIREMENT_EXCEPTION):
+            # We need to inform the leader that this is a CWL workflow problem
+            # and it needs to inform its caller.
+            failure_exit_code = CWL_UNSUPPORTED_REQUIREMENT_EXIT_CODE
         AbstractFileStore._terminateEvent.set()
-    
+
     ##########################################
     #Wait for the asynchronous chain of writes/updates to finish
-    ########################################## 
-    
-    blockFn() 
-    
     ##########################################
-    #All the asynchronous worker/update threads must be finished now, 
+
+    blockFn()
+
+    ##########################################
+    #All the asynchronous worker/update threads must be finished now,
     #so safe to test if they completed okay
-    ########################################## 
-    
-    if AbstractFileStore._terminateEvent.isSet():
+    ##########################################
+
+    if AbstractFileStore._terminateEvent.is_set():
         # Something has gone wrong.
-        
+
         # Clobber any garbage state we have for this job from failing with
         # whatever good state is still stored in the JobStore
         jobDesc = jobStore.load(jobStoreID)
@@ -583,38 +595,52 @@ def workerScript(jobStore, config, jobName, jobStoreID, redirectOutputToLogFile=
         statsDict.logs.messages = logMessages
 
     if (debugging or config.stats or statsDict.workers.logsToMaster) and not jobAttemptFailed:  # We have stats/logging to report back
-        jobStore.writeStatsAndLogging(json.dumps(statsDict, ensure_ascii=True).encode())
+        jobStore.writeStatsAndLogging(json.dumps(statsDict, ensure_ascii=True))
 
-    #Remove the temp dir
+    # Remove the temp dir
     cleanUp = config.cleanWorkDir
     if cleanUp == 'always' or (cleanUp == 'onSuccess' and not jobAttemptFailed) or (cleanUp == 'onError' and jobAttemptFailed):
-        shutil.rmtree(localWorkerTempDir)
-    
-    #This must happen after the log file is done with, else there is no place to put the log
+        def make_parent_writable(func: Callable[[str], Any], path: str, _: Any) -> None:
+            """
+            When encountering an error removing a file or directory, make sure
+            the parent directory is writable.
+
+            cwltool likes to lock down directory permissions, and doesn't clean
+            up after itself.
+            """
+            # Just chmod it for rwx for user. This can't work anyway if it isn't ours.
+            try:
+                os.chmod(os.path.dirname(path),  stat.S_IRUSR | stat.S_IWUSR | stat.S_IXUSR)
+            except PermissionError as e:
+                logger.error('Could not set permissions on %s to allow cleanup of %s: %s', os.path.dirname(path), path, e)
+                pass
+        shutil.rmtree(localWorkerTempDir, onerror=make_parent_writable)
+
+    # This must happen after the log file is done with, else there is no place to put the log
     if (not jobAttemptFailed) and jobDesc.command == None and next(jobDesc.successorsAndServiceHosts(), None) is None:
         # We can now safely get rid of the JobDescription, and all jobs it chained up
         for otherID in jobDesc.jobsToDelete:
             jobStore.delete(otherID)
         jobStore.delete(jobDesc.jobStoreID)
-        
+
     if jobAttemptFailed:
-        return 1
+        return failure_exit_code
     else:
         return 0
-        
-def parse_args(args):
+
+def parse_args(args: List[str]) -> argparse.Namespace:
     """
     Parse command-line arguments to the worker.
     """
-    
+
     # Drop the program name
     args = args[1:]
-    
+
     # Make the parser
     parser = argparse.ArgumentParser()
-    
+
     # Now add all the options to it
-    
+
     # Base required job information
     parser.add_argument("jobName", type=str,
         help="Text name of the job being run")
@@ -622,7 +648,7 @@ def parse_args(args):
         help="Information required to connect to the job store")
     parser.add_argument("jobStoreID", type=str,
         help="ID of the job within the job store")
-    
+
     # Additional worker abilities
     parser.add_argument("--context", default=[], action="append",
         help="""Pickled, base64-encoded context manager(s) to run job inside of.
@@ -630,41 +656,42 @@ def parse_args(args):
                 batch system, in the form of pickled Python context manager objects,
                 that the worker can then run before/after the job on the batch
                 system's behalf.""")
-   
+
     return parser.parse_args(args)
-    
-    
+
+
 @contextmanager
-def in_contexts(contexts):
+def in_contexts(contexts: List[str]) -> Iterator[None]:
     """
     Unpickle and enter all the pickled, base64-encoded context managers in the
     given list. Then do the body, then leave them all.
     """
-    
+
     if len(contexts) == 0:
         # Base case: nothing to do
         yield
     else:
         first = contexts[0]
         rest = contexts[1:]
-        
+
         try:
             manager = pickle.loads(base64.b64decode(first.encode('utf-8')))
         except:
             exc_info = sys.exc_info()
             logger.error('Exception while unpickling context manager: ', exc_info=exc_info)
             raise
-        
+
         with manager:
             # After entering this first context manager, do the rest.
             # It might set up stuff so we can decode later ones.
             with in_contexts(rest):
                 yield
 
-def main(argv=None):
+
+def main(argv: Optional[List[str]] = None) -> None:
     if argv is None:
         argv = sys.argv
-        
+
     # Parse our command line
     options = parse_args(argv)
 
@@ -679,10 +706,10 @@ def main(argv=None):
 
     jobStore = Toil.resumeJobStore(options.jobStoreLocator)
     config = jobStore.config
-    
+
     with in_contexts(options.context):
         # Call the worker
         exit_code = workerScript(jobStore, config, options.jobName, options.jobStoreID)
-    
+
     # Exit with its return value
     sys.exit(exit_code)
